﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Globalization;
using CodeStage.AntiCheat.ObscuredTypes;
using JsonDiffPatchDotNet;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using RocketCoroutine;
using RocketLog;
using RocketUtils;
using UnityEngine;

namespace RocketGM.PanStats
{
    public static class PlayerStats
    {
        private static Dictionary<string, BaseStat> _stats;
        private static List<string> _populatedStats;
        private static JsonSerializerSettings _serializerSettings;

        private static bool _playerDataWillBeOverridenByRemote;
        private static JsonDiffPatch _jsonDiffPatch = new JsonDiffPatch();

        private static readonly RocLog Log = new RocLog(nameof(PlayerStats), DebugLevels.Debug);

        public static void Initialize(Dictionary<string, BaseStat> stats, JsonSerializerSettings jsonSerializerSettings)
        {
            _playerDataWillBeOverridenByRemote = false;
            RocGmSessionReporting.PlayerDataIsDirty += OnPlayerDataIsDirty;

            _serializerSettings = jsonSerializerSettings;
            _stats = stats;
            _populatedStats = new List<string>();

            if (RocGm.GameSettings.RemotePlayerDataEnabled)
            {
                Log.Debug("RemotePlayerDataEnabled is true, waiting for initializing to GM.", "InitializePlayerStats");

                //Wait for initialization to get the player data from remote
                RocGm.DoAfterInitialized((_, _) =>
                {
                    //After initializing the game there is a chance that remote data sync is disabled.
                    if (RocGm.GameSettings.RemotePlayerDataEnabled)
                    {
                        Log.Debug("RemotePlayerDataEnabled is true after initializing to GM.", "InitializePlayerStats");

                        if (RocGm.IsRemotePlayerDataEmpty)
                        {
                            Log.Debug("RocGM.RemotePlayerData is empty, initializing from local data...", "InitializePlayerStats");
                            InitializeFromLocal();
                        }
                        else
                        {
                            Log.Debug("RocGM.RemotePlayerData is NOT empty, initializing from remote data...", "InitializePlayerStats");
                            InitializeFromRemoteData();
                            InitializeFromLocal();
                        }
                    }
                    else
                    {
                        Log.Debug("RemotePlayerDataEnabled is false after initializing to GM.", "InitializePlayerStats");
                        InitializeFromLocal();
                    }

                    CoroutineController.StartCoroutine(SavePlayerStatsCoroutine(15));
                });
            }
            else
            {
                Log.Debug("RemotePlayerDataEnabled is false, do NOT wait for initialization, initialize from local.", "InitializePlayerStats");

                InitializeFromLocal();

                CoroutineController.StartCoroutine(SavePlayerStatsCoroutine(15));
            }

        }
        private static void OnPlayerDataIsDirty()
        {
            Log.Warning("Player data will be overriden by remote! Clearing local stats", nameof(OnPlayerDataIsDirty));
            _playerDataWillBeOverridenByRemote = true;

            foreach (KeyValuePair<string, BaseStat> keyValuePair in _stats)
            {
                var stat = keyValuePair.Value;
                if (!stat.SyncRemote)
                {
                    ObscuredPrefs.DeleteKey(keyValuePair.Key);
                    Log.Warning($"Stat with key {keyValuePair.Key} will be cleared.", nameof(OnPlayerDataIsDirty));
                }
            }
        }

        private static void InitializeFromRemoteData()
        {
            try
            {
                var unsentSessionDataContainer = RocGmSessionReporting.GetSentSessionDataContainer();
                if (unsentSessionDataContainer.SentImpressionDataList != null)
                {
                    unsentSessionDataContainer.SentImpressionDataList.Sort((a, b) => a.RemotePlayerDataCount.CompareTo(b.RemotePlayerDataCount));
                    foreach (var sentSessionData in unsentSessionDataContainer.SentImpressionDataList)
                    {
                        //Diff data will be cleared if player data is tagged with dirty from gm
                        if (sentSessionData.RemotePlayerDataDiff != "")
                        {
                            RocGm.RemotePlayerData = _jsonDiffPatch.Patch(RocGm.RemotePlayerData, sentSessionData.RemotePlayerDataDiff);
                        }
                    }
                }

                ParseJsonStringAndSaveToPrefs(RocGm.RemotePlayerData);
            }
            catch (Exception e)
            {
                Log.Exception(e.Message, nameof(InitializeFromRemoteData));
                RocGm.RemotePlayerData = RocGm.EmptyRemotePlayerData;
                _populatedStats.Clear();
                InitializeFromLocal();
            }
        }

        public static void ParseJsonStringAndSaveToPrefs(string json)
        {
            var deserializedRemoteKVP = JsonConvert.DeserializeObject<Dictionary<string, object>>(json);
            foreach (var kvp in deserializedRemoteKVP)
            {
                if (TryPopulateFromJToken(kvp.Key, kvp.Value, out var stat))
                {
                    _populatedStats.Add(kvp.Key);
                    SaveToPlayerPrefs(kvp.Key, stat.GetStatType());
                    stat.IsDirty = false;
                }
            }
        }

        private static void InitializeFromLocal()
        {
            var stats = new Dictionary<string, BaseStat>(_stats);
            foreach (KeyValuePair<string, BaseStat> keyValuePair in stats)
            {
                string key = keyValuePair.Key;
                BaseStat stat = keyValuePair.Value;

                if (_populatedStats.Contains(key) && stat.SyncRemote)
                    continue;

                if (ObscuredPrefs.HasKey(key))
                {
                    //Exists in player pref, update dictionary.
                    PopulateFromPlayerPrefs(key);
                }
                else
                {
                    //Does not exist in player pref, record it.
                    SaveToPlayerPrefs(key, stat.GetStatType());
                }
            }
        }

        private static bool TryPopulateFromJToken(string statKey, object statValue, out BaseStat stat)
        {
            var statValueString = statValue.ToString();
            stat = null;
            if (_stats.ContainsKey(statKey))
            {
                stat = _stats[statKey];

                if (!stat.SyncRemote)
                    return false;

                Type statType = stat.GetStatType();

                Log.Debug("Populating " + statKey);

                if (statType == typeof(int))
                {
                    int value = LoadValueFromJsonString<int>(statValueString);
                    IntStat intStat = (IntStat)_stats[statKey];

                    _stats[statKey] = new IntStat(intStat.Min, intStat.Max, value, stat.SyncRemote);

                    Log.Debug($"Value {value} with Key {statKey} Loaded", "PopulateFromPlayerPrefs");
                }
                else if (statType == typeof(float))
                {
                    float value = LoadValueFromJsonString<float>(statValueString);
                    FloatStat floatStat = (FloatStat)_stats[statKey];

                    _stats[statKey] = new FloatStat(floatStat.Min, floatStat.Max, value, stat.SyncRemote);

                    Log.Debug($"Value {value} with Key {statKey} Loaded", "PopulateFromPlayerPrefs");
                }
                else if (statType == typeof(string))
                {
                    string value = LoadValueFromJsonString<string>(statValueString);

                    _stats[statKey] = new StringStat(value, stat.SyncRemote);

                    Log.Debug($"Value {value} with Key {statKey} Loaded", "PopulateFromPlayerPrefs");
                }
                else if (statType.Name.Contains("ObservableList") || statType.Name.Contains("ObservableDict") || statType == typeof(IObservableModel))
                {
                    string value = LoadValueFromJsonString<string>(statValueString);
                    var stringStat = (StringStat)_stats[statKey];
                    stringStat.Reconstruct(value, stringStat.SyncRemote);
                    Log.Debug($"Value {value} with Key {statKey} Loaded", "PopulateFromPlayerPrefs");
                }
                else if (statType == typeof(bool))
                {
                    bool value = LoadValueFromJsonString<bool>(statValueString);

                    _stats[statKey] = new BoolStat(value, stat.SyncRemote);

                    Log.Debug($"Value {value} with Key {statKey} Loaded", "PopulateFromPlayerPrefs");
                }
                else if (statType == typeof(long))
                {
                    long value = LoadValueFromJsonString<long>(statValueString);
                    LongStat longStat = (LongStat)_stats[statKey];

                    _stats[statKey] = new LongStat(longStat.Min, longStat.Max, value, stat.SyncRemote);

                    Log.Debug($"Value {value} with key {statKey} Loaded", "PopulateFromPlayerPrefs");
                }
                else if (statType == typeof(DateTime))
                {
                    long value = LoadValueFromJsonString<long>(statValueString);
                    DateTime date;
                    try
                    {
                        date = new DateTime(value);

                    }
                    catch (Exception e)
                    {
                        Log.Debug("Incoming value was saved as Binary", "PopulateFromPlayerPrefs");
                        date = DateTime.FromBinary(value);
                    }

                    _stats[statKey] = new DateStat(value, stat.SyncRemote);

                    try
                    {
                        Log.Debug($"Value {date.ToString("G", CultureInfo.InvariantCulture)} with key {statKey} Loaded", "PopulateFromPlayerPrefs");
                    }
                    catch
                    {
                        Log.Warning($"{statKey} can not be represented as string with current language info. Culture: {CultureInfo.CurrentCulture.Name}");
                    }
                }
                else if (statType == typeof(byte))
                {
                    byte value = LoadValueFromJsonString<byte>(statValueString);
                    ByteStat byteStat = (ByteStat)_stats[statKey];

                    _stats[statKey] = new ByteStat(byteStat.Min, byteStat.Max, value, stat.SyncRemote);

                    Log.Debug($"Value {value} with Key {statKey} Loaded", "PopulateFromPlayerPrefs");
                }
                else if (statType == typeof(double))
                {
                    double value = LoadValueFromJsonString<double>(statValueString);
                    DoubleStat doubleStat = (DoubleStat)_stats[statKey];

                    _stats[statKey] = new DoubleStat(doubleStat.Min, doubleStat.Max, value, stat.SyncRemote);

                    Log.Debug($"Value {value} with Key {statKey} Loaded", "PopulateFromPlayerPrefs");
                }
                else if (statType == typeof(short))
                {
                    short value = LoadValueFromJsonString<short>(statValueString);
                    ShortStat shortStat = (ShortStat)_stats[statKey];

                    _stats[statKey] = new ShortStat(shortStat.Min, shortStat.Max, value, stat.SyncRemote);

                    Log.Debug($"Value {value} with Key {statKey} Loaded", "PopulateFromPlayerPrefs");
                }
                else if (statType == typeof(ushort))
                {
                    ushort value = LoadValueFromJsonString<ushort>(statValueString);
                    UShortStat uShortStat = (UShortStat)_stats[statKey];

                    _stats[statKey] = new UShortStat(uShortStat.Min, uShortStat.Max, value, stat.SyncRemote);

                    Log.Debug($"Value {value} with Key {statKey} Loaded", "PopulateFromPlayerPrefs");
                }
                else if (statType == typeof(uint))
                {
                    uint value = LoadValueFromJsonString<uint>(statValueString);
                    UIntStat uShortStat = (UIntStat)_stats[statKey];

                    _stats[statKey] = new UIntStat(uShortStat.Min, uShortStat.Max, value, stat.SyncRemote);

                    Log.Debug($"Value {value} with Key {statKey} Loaded", "PopulateFromPlayerPrefs");
                }
                else
                {
                    throw new InvalidCastException($"Wrong Type {statType.Name} in default player data dictionary ");
                }

                return true;
            }

            return false;
        }

        private static void PopulateFromPlayerPrefs(string statKey)
        {
            var stat = _stats[statKey];

            Type statType = stat.GetStatType();

            Log.Debug("Populating " + statKey);

            if (statType == typeof(int))
            {
                int value = LoadValueFromPlayerPrefs<int>(statKey);
                IntStat intStat = (IntStat)_stats[statKey];

                _stats[statKey] = new IntStat(intStat.Min, intStat.Max, value, stat.SyncRemote);

                Log.Debug($"Value {value} with Key {statKey} Loaded", "PopulateFromPlayerPrefs");
            }
            else if (statType == typeof(float))
            {
                float value = LoadValueFromPlayerPrefs<float>(statKey);
                FloatStat floatStat = (FloatStat)_stats[statKey];

                _stats[statKey] = new FloatStat(floatStat.Min, floatStat.Max, value, stat.SyncRemote);

                Log.Debug($"Value {value} with Key {statKey} Loaded", "PopulateFromPlayerPrefs");
            }
            else if (statType == typeof(string))
            {
                string value = LoadValueFromPlayerPrefs<string>(statKey);

                _stats[statKey] = new StringStat(value, stat.SyncRemote);

                Log.Debug($"Value {value} with Key {statKey} Loaded", "PopulateFromPlayerPrefs");
            }
            else if (statType.Name.Contains("ObservableList") || statType.Name.Contains("ObservableDict") || statType == typeof(IObservableModel))
            {
                string value = LoadValueFromPlayerPrefs<string>(statKey);
                var stringStat = (StringStat)_stats[statKey];
                stringStat.Reconstruct(value, stringStat.SyncRemote);
                Log.Debug($"Value {value} with Key {statKey} Loaded", "PopulateFromPlayerPrefs");
            }
            else if (statType == typeof(bool))
            {
                bool value = LoadValueFromPlayerPrefs<bool>(statKey);

                _stats[statKey] = new BoolStat(value, stat.SyncRemote);

                Log.Debug($"Value {value} with Key {statKey} Loaded", "PopulateFromPlayerPrefs");
            }
            else if (statType == typeof(long))
            {
                long value = LoadValueFromPlayerPrefs<long>(statKey);
                LongStat longStat = (LongStat)_stats[statKey];

                _stats[statKey] = new LongStat(longStat.Min, longStat.Max, value, stat.SyncRemote);

                Log.Debug($"Value {value} with key {statKey} Loaded", "PopulateFromPlayerPrefs");
            }
            else if (statType == typeof(DateTime))
            {
                long value = LoadValueFromPlayerPrefs<long>(statKey);
                DateTime date;
                try
                {
                    date = new DateTime(value);

                }
                catch (Exception e)
                {
                    Log.Debug("Incoming value was saved as Binary", "PopulateFromPlayerPrefs");
                    date = DateTime.FromBinary(value);
                }

                _stats[statKey] = new DateStat(value, stat.SyncRemote);

                try
                {
                    Log.Debug($"Value {date.ToString("G", CultureInfo.InvariantCulture)} with key {statKey} Loaded", "PopulateFromPlayerPrefs");
                }
                catch
                {
                    Log.Warning($"{statKey} can not be represented as string with current language info. Culture: {CultureInfo.CurrentCulture.Name}");
                }
            }
            else if (statType == typeof(byte))
            {
                byte value = LoadValueFromPlayerPrefs<byte>(statKey);
                ByteStat byteStat = (ByteStat)_stats[statKey];

                _stats[statKey] = new ByteStat(byteStat.Min, byteStat.Max, value, stat.SyncRemote);

                Log.Debug($"Value {value} with Key {statKey} Loaded", "PopulateFromPlayerPrefs");
            }
            else if (statType == typeof(double))
            {
                double value = LoadValueFromPlayerPrefs<double>(statKey);
                DoubleStat doubleStat = (DoubleStat)_stats[statKey];

                _stats[statKey] = new DoubleStat(doubleStat.Min, doubleStat.Max, value, stat.SyncRemote);

                Log.Debug($"Value {value} with Key {statKey} Loaded", "PopulateFromPlayerPrefs");
            }
            else if (statType == typeof(short))
            {
                short value = LoadValueFromPlayerPrefs<short>(statKey);
                ShortStat shortStat = (ShortStat)_stats[statKey];

                _stats[statKey] = new ShortStat(shortStat.Min, shortStat.Max, value, stat.SyncRemote);

                Log.Debug($"Value {value} with Key {statKey} Loaded", "PopulateFromPlayerPrefs");
            }
            else if (statType == typeof(ushort))
            {
                ushort value = LoadValueFromPlayerPrefs<ushort>(statKey);
                UShortStat uShortStat = (UShortStat)_stats[statKey];

                _stats[statKey] = new UShortStat(uShortStat.Min, uShortStat.Max, value, stat.SyncRemote);

                Log.Debug($"Value {value} with Key {statKey} Loaded", "PopulateFromPlayerPrefs");
            }
            else if (statType == typeof(uint))
            {
                uint value = LoadValueFromPlayerPrefs<uint>(statKey);
                UIntStat uShortStat = (UIntStat)_stats[statKey];

                _stats[statKey] = new UIntStat(uShortStat.Min, uShortStat.Max, value, stat.SyncRemote);

                Log.Debug($"Value {value} with Key {statKey} Loaded", "PopulateFromPlayerPrefs");
            }
            else
            {
                throw new InvalidCastException($"Wrong Type {statType.Name} in default player data dictionary ");
            }
        }

        public static bool HasStat(string statKey)
        {
            return _stats.ContainsKey(statKey);
        }

        public static T Get<T>(string statKey)
        {
            try
            {
                return _stats[statKey].Get<T>();
            }
            catch (Exception e)
            {
                Log.Warning(statKey + " does NOT exist in DefaultPlayerStats.");
                throw;
            }

        }

        public static T GetMin<T>(string statKey)
        {
            if (typeof(T) == typeof(int))
            {
                return (T)Convert.ChangeType(((IntStat)_stats[statKey]).Min, typeof(T));
            }
            if (typeof(T) == typeof(float))
            {
                return (T)Convert.ChangeType(((FloatStat)_stats[statKey]).Min, typeof(T));
            }
            if (typeof(T) == typeof(long))
            {
                return (T)Convert.ChangeType(((LongStat)_stats[statKey]).Min, typeof(T));
            }
            if (typeof(T) == typeof(byte))
            {
                return (T)Convert.ChangeType(((ByteStat)_stats[statKey]).Min, typeof(T));
            }
            if (typeof(T) == typeof(double))
            {
                return (T)Convert.ChangeType(((DoubleStat)_stats[statKey]).Min, typeof(T));
            }
            if (typeof(T) == typeof(short))
            {
                return (T)Convert.ChangeType(((ShortStat)_stats[statKey]).Min, typeof(T));
            }
            if (typeof(T) == typeof(ushort))
            {
                return (T)Convert.ChangeType(((UShortStat)_stats[statKey]).Min, typeof(T));
            }
            if (typeof(T) == typeof(uint))
            {
                return (T)Convert.ChangeType(((UIntStat)_stats[statKey]).Min, typeof(T));
            }
            throw new CustomStatException(typeof(T));
        }

        public static T GetMax<T>(string statKey)
        {
            if (typeof(T) == typeof(int))
            {
                return (T)Convert.ChangeType(((IntStat)_stats[statKey]).Max, typeof(T));
            }
            if (typeof(T) == typeof(float))
            {
                return (T)Convert.ChangeType(((FloatStat)_stats[statKey]).Max, typeof(T));
            }
            if (typeof(T) == typeof(long))
            {
                return (T)Convert.ChangeType(((LongStat)_stats[statKey]).Max, typeof(T));
            }
            if (typeof(T) == typeof(byte))
            {
                return (T)Convert.ChangeType(((ByteStat)_stats[statKey]).Max, typeof(T));
            }
            if (typeof(T) == typeof(double))
            {
                return (T)Convert.ChangeType(((DoubleStat)_stats[statKey]).Max, typeof(T));
            }
            if (typeof(T) == typeof(short))
            {
                return (T)Convert.ChangeType(((ShortStat)_stats[statKey]).Max, typeof(T));
            }
            if (typeof(T) == typeof(ushort))
            {
                return (T)Convert.ChangeType(((UShortStat)_stats[statKey]).Max, typeof(T));
            }
            if (typeof(T) == typeof(uint))
            {
                return (T)Convert.ChangeType(((UIntStat)_stats[statKey]).Max, typeof(T));
            }
            throw new CustomStatException(typeof(T));
        }

        public static void Set<T>(string statKey, T value)
        {
            try
            {
                _stats[statKey].Set(value);
            }
            catch (Exception e)
            {
                Log.Warning(statKey + " does NOT exist in DefaultPlayerStats.");
                throw;
            }
        }

        private static T LoadValueFromPlayerPrefs<T>(string statKey)
        {
            if (typeof(T) == typeof(int))
            {
                return (T)Convert.ChangeType(ObscuredPrefs.GetInt(statKey), typeof(T));
            }
            if (typeof(T) == typeof(float))
            {
                return (T)Convert.ChangeType(ObscuredPrefs.GetFloat(statKey), typeof(T));
            }
            if (typeof(T) == typeof(string))
            {
                return (T)Convert.ChangeType(ObscuredPrefs.GetString(statKey), typeof(T));
            }
            if (typeof(T) == typeof(bool))
            {
                return (T)Convert.ChangeType(ObscuredPrefs.GetBool(statKey), typeof(T));
            }
            if (typeof(T) == typeof(long))
            {
                return (T)Convert.ChangeType(ObscuredPrefs.GetLong(statKey), typeof(T));
            }
            if (typeof(T) == typeof(byte))
            {
                return (T)Convert.ChangeType(Convert.ToByte(ObscuredPrefs.GetInt(statKey)), typeof(T));
            }
            if (typeof(T) == typeof(double))
            {
                return (T)Convert.ChangeType(ObscuredPrefs.GetDouble(statKey), typeof(T));
            }
            if (typeof(T) == typeof(short))
            {
                return (T)Convert.ChangeType(Convert.ToInt16(ObscuredPrefs.GetInt(statKey)), typeof(T));
            }
            if (typeof(T) == typeof(ushort))
            {
                return (T)Convert.ChangeType(Convert.ToUInt16(ObscuredPrefs.GetInt(statKey)), typeof(T));
            }
            if (typeof(T) == typeof(uint))
            {
                return (T)Convert.ChangeType(Convert.ToUInt32(ObscuredPrefs.GetInt(statKey)), typeof(T));
            }
            throw new InvalidCastException();
        }

        private static T LoadValueFromJsonString<T>(string jsonString)
        {
            if (typeof(T) == typeof(int))
            {
                return (T)Convert.ChangeType(Convert.ToInt32(jsonString), typeof(T));
            }
            if (typeof(T) == typeof(float))
            {
                return (T)Convert.ChangeType(Convert.ToSingle(jsonString), typeof(T));
            }
            if (typeof(T) == typeof(string))
            {
                return (T)Convert.ChangeType(jsonString, typeof(T));
            }
            if (typeof(T) == typeof(bool))
            {
                return (T)Convert.ChangeType(Convert.ToBoolean(jsonString), typeof(T));
            }
            if (typeof(T) == typeof(long))
            {
                return (T)Convert.ChangeType(Convert.ToInt64(jsonString), typeof(T));
            }
            if (typeof(T) == typeof(byte))
            {
                return (T)Convert.ChangeType(Convert.ToByte(jsonString), typeof(T));
            }
            if (typeof(T) == typeof(double))
            {
                return (T)Convert.ChangeType(Convert.ToDouble(jsonString), typeof(T));
            }
            if (typeof(T) == typeof(short))
            {
                return (T)Convert.ChangeType(Convert.ToInt16(jsonString), typeof(T));
            }
            if (typeof(T) == typeof(ushort))
            {
                return (T)Convert.ChangeType(Convert.ToUInt16(jsonString), typeof(T));
            }
            if (typeof(T) == typeof(uint))
            {
                return (T)Convert.ChangeType(Convert.ToUInt32(jsonString), typeof(T));
            }
            throw new InvalidCastException();
        }

        private static void SaveToPlayerPrefs(string statKey, Type type)
        {
            if (type == typeof(int))
            {
                ObscuredPrefs.SetInt(statKey, Get<int>(statKey));
                Log.Debug($"Value {Get<int>(statKey)} with Key {statKey} Saved", "SaveToPlayerPrefs");
            }
            else if (type == typeof(float))
            {
                ObscuredPrefs.SetFloat(statKey, Get<float>(statKey));
                Log.Debug($"Value {Get<float>(statKey)} with Key {statKey} Saved", "SaveToPlayerPrefs");
            }
            else if (type == typeof(string))
            {
                ObscuredPrefs.SetString(statKey, Get<string>(statKey));
                Log.Debug($"Value {Get<string>(statKey)} with Key {statKey} Saved", "SaveToPlayerPrefs");
            }
            else if (type.Name.Contains("ObservableList") || type.Name.Contains("ObservableDict") || type == typeof(IObservableModel))
            {
                var stat = _stats[statKey] as StringStat;
                if (stat != null)
                    ObscuredPrefs.SetString(statKey, stat.Current);
            }
            else if (type == typeof(bool))
            {
                ObscuredPrefs.SetBool(statKey, Get<bool>(statKey));
                Log.Debug($"Value {Get<bool>(statKey)} with Key {statKey} Saved", "SaveToPlayerPrefs");
            }
            else if (type == typeof(long))
            {
                ObscuredPrefs.SetLong(statKey, Get<long>(statKey));
                Log.Debug($"Value {Get<long>(statKey)} with Key {statKey} Saved", "SaveToPlayerPrefs");
            }
            else if (_stats[statKey].GetStatType() == typeof(DateTime))
            {
                ObscuredPrefs.SetLong(statKey, Get<DateTime>(statKey).Ticks);

                try
                {
                    Log.Debug(
                        $"Value {Get<DateTime>(statKey).ToString("G", CultureInfo.InvariantCulture)} with Key {statKey} Saved",
                        "SaveToPlayerPrefs");
                }
                catch
                {
                    Log.Warning($"{statKey} can not be represented as string with current language info. Culture: {CultureInfo.CurrentCulture.Name}");
                }
            }
            else if (type == typeof(byte))
            {
                ObscuredPrefs.SetInt(statKey, Get<byte>(statKey));
                Log.Debug($"Value {Get<byte>(statKey)} with Key {statKey} Saved", "SaveToPlayerPrefs");
            }
            else if (type == typeof(double))
            {
                ObscuredPrefs.SetDouble(statKey, Get<double>(statKey));
                Log.Debug($"Value {Get<double>(statKey)} with Key {statKey} Saved", "SaveToPlayerPrefs");
            }
            else if (type == typeof(short))
            {
                ObscuredPrefs.SetInt(statKey, Get<short>(statKey));
                Log.Debug($"Value {Get<short>(statKey)} with Key {statKey} Saved", "SaveToPlayerPrefs");
            }
            else if (type == typeof(ushort))
            {
                ObscuredPrefs.SetInt(statKey, Get<ushort>(statKey));
                Log.Debug($"Value {Get<ushort>(statKey)} with Key {statKey} Saved", "SaveToPlayerPrefs");
            }
            else if (type == typeof(uint))
            {
                ObscuredPrefs.SetUInt(statKey, Get<uint>(statKey));
                Log.Debug($"Value {Get<uint>(statKey)} with Key {statKey} Saved", "SaveToPlayerPrefs");
            }
            else
            {
                throw new InvalidCastException();
            }
        }

        private static IEnumerator SavePlayerStatsCoroutine(int waitTime)
        {
            while (true)
            {
                if (_playerDataWillBeOverridenByRemote)
                    yield break;

                SaveDirtyStats();
                yield return new WaitForSecondsRealtime(waitTime);
            }
            // ReSharper disable once IteratorNeverReturns
        }

        public static void SaveDirtyStats()
        {
            if (_stats == null)
            {
                Log.Warning("Player stats is not initialized");
                return;
            }
            foreach (var stat in _stats)
            {
                if (stat.Value.IsDirty)
                {
                    SaveToPlayerPrefs(stat.Key, stat.Value.GetStatType());
                    stat.Value.IsDirty = false;
                    Log.Debug($" {stat.Key} stat saved as {stat.Value.GetStatType()}", "SaveTimerCoroutine");
                }
            }
        }

        public static bool TryGetPlayerDataAsJsonString(bool skipLocalData, out string jsonString)
        {
            if (_stats == null)
            {
                Log.Error("Player Stats is NOT initialized yet!", nameof(TryGetPlayerDataAsJsonString));
                jsonString = null;
                return false;
            }

            var stats = new Dictionary<string, BaseStat>();
            foreach (var (key, value) in _stats)
            {
                if (skipLocalData && !value.SyncRemote)
                    continue;
                stats.Add(key, value);
            }

            jsonString = JsonConvert.SerializeObject(stats, _serializerSettings);
            return true;
        }

        private static void OnAppQuit()
        {
            SaveDirtyStats();
        }

        private static void OnPauseStateChanged(bool isPaused)
        {
            if (isPaused)
            {
                SaveDirtyStats();
            }
        }
    }
}
