﻿using System;
using RocketLog;

namespace RocketGM.RemoteSettings
{
    public abstract class RemoteVariable
    {
        public event Action ValueUpdated;
        public bool IsUpdated;

        protected string RemoteKey;
        protected object DefaultValue;
        protected object Value;

        private readonly RocLog _log = new RocLog(nameof(RemoteVariable), DebugLevels.Info);

        protected RemoteVariable(string remoteKey, object defaultValue)
        {
            RemoteKey = remoteKey;
            DefaultValue = defaultValue;
            Value = DefaultValue;

            if (!string.IsNullOrEmpty(RemoteKey))
            {
                //First try to get value if remote settings is initialized.
                TryUpdateValue();
                //Register for updates
                RocGm.RemoteSettings.RemoteSettingsReceived += TryUpdateValue;
            }
        }

        ~RemoteVariable()
        {
            RocGm.RemoteSettings.RemoteSettingsReceived -= TryUpdateValue;
        }

        public static implicit operator string(RemoteVariable i)
        {
            return i.ConvertToString();
        }

        public override string ToString()
        {
            return ConvertToString();
        }

        private void TryUpdateValue(bool isSuccess = true)
        {
            if (!RocGm.RemoteSettings.IsRemoteSettingsReceived) return;

            if (!RocGm.RemoteSettings.HasKey(RemoteKey))
            {
                _log.Warning($"Key: '{RemoteKey}' does not exist in PanRemoteSettings.");
                return;
            }

            UpdateValue();

            IsUpdated = true;
            if (ValueUpdated != null) ValueUpdated.Invoke();
        }

        protected abstract string ConvertToString();
        protected abstract void UpdateValue();
    }

    public class RemoteInt : RemoteVariable
    {
        public RemoteInt(string remoteKey, int defaultValue) : base(remoteKey, defaultValue)
        {
        }

        public static implicit operator int(RemoteInt i)
        {
            return (int)i.Value;
        }

        protected override string ConvertToString()
        {
            return ((int)Value).ToString();
        }

        public int Default => (int)DefaultValue;

        protected override void UpdateValue()
        {
            Value = RocGm.RemoteSettings.GetInt(RemoteKey);
        }
    }

    public class RemoteBool : RemoteVariable
    {
        public RemoteBool(string remoteKey, bool defaultValue) : base(remoteKey, defaultValue)
        {
        }

        public static implicit operator bool(RemoteBool i)
        {
            return (bool)i.Value;
        }

        protected override string ConvertToString()
        {
            return ((bool)Value).ToString();
        }

        public bool Default => (bool)DefaultValue;

        protected override void UpdateValue()
        {
            Value = RocGm.RemoteSettings.GetBool(RemoteKey);
        }
    }

    public class RemoteString : RemoteVariable
    {
        public RemoteString(string remoteKey, string defaultValue) : base(remoteKey, defaultValue)
        {
        }

        public static implicit operator string(RemoteString i)
        {
            return (string)i.Value;
        }

        protected override string ConvertToString()
        {
            return (string)Value;
        }

        public string Default => (string)DefaultValue;

        protected override void UpdateValue()
        {
            Value = RocGm.RemoteSettings.GetString(RemoteKey);
        }
    }

    public class RemoteJsonString : RemoteString
    {
        public RemoteJsonString(string remoteKey, string defaultValue) : base(remoteKey, defaultValue)
        {
        }
    }

    public class RemoteFloat : RemoteVariable
    {
        public RemoteFloat(string remoteKey, float defaultValue) : base(remoteKey, defaultValue)
        {
        }

        public static implicit operator float(RemoteFloat i)
        {
            return (float)i.Value;
        }

        protected override string ConvertToString()
        {
            return ((float)Value).ToString(System.Globalization.CultureInfo.InvariantCulture);
        }

        public float Default => (float)DefaultValue;

        protected override void UpdateValue()
        {
            Value = RocGm.RemoteSettings.GetFloat(RemoteKey);
        }
    }
}