﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using RocketUtils;
using UnityEngine;
using Object = System.Object;

namespace RocketGM.RemoteSettings
{
    public class RocGmRemoteSettings : RocRemoteSettingsProvider
    {
        private const string _remoteSettingsKey = "settings";

        private static Dictionary<String, Object> _settings = new Dictionary<string, object>();

        public RocGmRemoteSettings()
        {
            RocGm.Initialized += PanGmOnInitialized;
            RocGm.StatusChanged += PanGmOnStatusChanged;
        }

        private void PanGmOnInitialized(Dictionary<string, object> parameters, RocGmGameSettings gameSettings)
        {
            RocGm.Initialized -= PanGmOnInitialized;
            RocGm.StatusChanged -= PanGmOnStatusChanged;

            if (!parameters.ContainsKey(_remoteSettingsKey))
            {
                Log.Debug("Server did NOT return any RemoteSettings");
                return;
            }

            var remoteSettings = parameters[_remoteSettingsKey].ToString();
            if (RocGm.Config.IsEncryptedMessagingActive)
            {
                remoteSettings = RocGmSecurity.Decrypt(remoteSettings, RocGm.Config.SecurityKey);
            }

            _settings = Json.Deserialize(remoteSettings) as Dictionary<String, Object>;

            PlayerPrefs.SetString(_remoteSettingsKey, remoteSettings);
            SetRemoteSettingsReceived();
        }

        private void PanGmOnStatusChanged(RocGm.InitializationStatus initializationStatus)
        {
            switch (initializationStatus)
            {
                case RocGm.InitializationStatus.Failed:
                case RocGm.InitializationStatus.WaitingToConnectInternet:
                    RocGm.StatusChanged -= PanGmOnStatusChanged;

                    InitializeInOfflineMode();
                    break;
            }
        }


        private void InitializeInOfflineMode()
        {
            bool isSuccess = PlayerPrefs.HasKey(_remoteSettingsKey);

            if (isSuccess)
            {
                _settings = Json.Deserialize(PlayerPrefs.GetString(_remoteSettingsKey, "")) as Dictionary<String, Object>;
            }

            SetRemoteSettingsReceived(isSuccess);
        }


        public override bool HasKey(string key)
        {
            if (!IsRemoteSettingsReceived)
            {
                Log.Error("PanRemoteSettings did NOT received!");
                return false;
            }
            return _settings.ContainsKey(key);
        }

        public override int GetInt(string key, int defaultValue = 0)
        {
            int value = defaultValue;

            if (!IsRemoteSettingsReceived)
            {
                Log.Error("PanRemoteSettings did NOT received!");
                return value;
            }

            if (!_settings.ContainsKey(key)) return value;

            int.TryParse(_settings[key].ToString(), out value);

            return value;
        }

        public override float GetFloat(string key, float defaultValue = 0)
        {
            float value = defaultValue;

            if (!IsRemoteSettingsReceived)
            {
                Log.Error("PanRemoteSettings did NOT received!");
                return value;
            }

            if (!_settings.ContainsKey(key)) return value;

            float.TryParse(_settings[key].ToString(), out value);

            return value;
        }

        public override bool GetBool(string key, bool defaultValue = false)
        {
            bool value = defaultValue;

            if (!IsRemoteSettingsReceived)
            {
                Log.Error("PanRemoteSettings did NOT received!");
                return value;
            }

            if (!_settings.ContainsKey(key)) return value;

            bool.TryParse(_settings[key].ToString(), out value);

            return value;
        }

        public override string GetString(string key, string defaultValue = "")
        {
            string value = defaultValue;

            if (!IsRemoteSettingsReceived)
            {
                Log.Error("PanRemoteSettings did NOT received!");
                return value;
            }

            if (!_settings.ContainsKey(key)) return value;

            value = _settings[key].ToString();

            return value;
        }

        public override List<T> GetArray<T>(string key, List<T> defaultValue)
        {
            defaultValue = new List<T>(defaultValue);

            if (!IsRemoteSettingsReceived)
            {
                Log.Error("PanRemoteSettings did NOT received!");
                return defaultValue;
            }

            if (!_settings.ContainsKey(key)) return defaultValue;

            var deserializedList = _settings[key] as IList;
            if (deserializedList == null)
            {
                Log.Error("Incoming object is NOT list!!!!");
                return defaultValue;
            }

            var values = new List<T>();

            foreach (var value in deserializedList)
            {
                try
                {
                    var newValue = (T)Convert.ChangeType(value, typeof(T));
                    values.Add(newValue);
                }
                catch
                {
                }
            }

            return values.Count == 0 ? defaultValue : values;
        }
    }
}
