﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Globalization;
using System.Threading.Tasks;
using Newtonsoft.Json.Linq;
using RocketCoroutine;
using RocketGM.Models.UpdateSession;
using RocketGM.RemoteSettings;
using RocketGM.Utils;
using RocketLog;
using RocketUtils;
using UnityEngine;
using UnityEngine.Networking;
using Object = System.Object;

namespace RocketGM
{
    /// <summary>
    /// Used for serialization of server time that is returned by GM endpoint.
    /// </summary>
    [Serializable]
    public class ServerDateTime
    {
        public string timestamp;
    }

    public class RocGm
    {
        private sealed class WaitUntilWithTimeout : CustomYieldInstruction
        {
            private readonly Func<bool> _predicate;
            private float _timeout;
            private bool WaitForDoneProcess()
            {
                _timeout -= Time.deltaTime;
                return _timeout <= 0f || _predicate();
            }

            public override bool keepWaiting => !WaitForDoneProcess();

            public WaitUntilWithTimeout(float timeout, Func<bool> predicate)
            {
                _predicate = predicate;
                _timeout = timeout;
            }
        }

        public enum InitializationStatus
        {
            Failed = -1,
            NotStarted = 0,
            Initializing = 1,
            RetryingToSolveConnectionProblems = 2,
            WaitingToConnectInternet = 3,
            Initialized = 5,
        }

        public enum ServerTimeStatus
        {
            Failed = -1,
            NotStarted = 0,
            Initializing = 1,
            Initialized = 5
        }

        public static event Action<Dictionary<string, object>, RocGmGameSettings> Initialized;
        public static event Action<InitializationStatus> StatusChanged;
        public static event Action<ServerTimeStatus> FetchServerTimeStatusChanged;
        public static event Action SentPlayerPropertyToGameMaster;

        public static RocGmReporting Reporting;
        public static RocGmLiveTune LiveTune;

        private static RocRemoteSettingsProvider _remoteSettings;
        public static RocRemoteSettingsProvider RemoteSettings
        {
            get
            {
                if (_remoteSettings == null)
                {
                    _remoteSettings = new RocGmRemoteSettings();
                }

                return _remoteSettings;
            }
        }
        public static RocGmPlayerProgress PlayerProgress;

        public static long PlayerId;

        public static bool IsInitialized
        {
            get { return PlayerId != 0; }
        }


        private static InitializationStatus _status = InitializationStatus.NotStarted;

        public static InitializationStatus Status
        {
            get { return _status; }
            set
            {
                if (_status == value)
                    return;

                _status = value;

                if (StatusChanged != null) StatusChanged.Invoke(_status);
            }
        }

        private static ServerTimeStatus _fetchServerTimeStatus = ServerTimeStatus.NotStarted;

        public static ServerTimeStatus FetchServerTimeStatus
        {
            get => _fetchServerTimeStatus;
            set
            {
                if (_fetchServerTimeStatus == value)
                    return;

                _fetchServerTimeStatus = value;

                FetchServerTimeStatusChanged?.Invoke(_fetchServerTimeStatus);
            }
        }

        public static bool IsMaintenance { get; private set; }

        private static bool _failedOnce;

        private static Dictionary<string, object> _parameters;
        public static RocGmGameSettings GameSettings;
        public static RocGmPlayerProperties PlayerProperties;

        #region Time

        private static DateTime _minUtcDateTime = new DateTime(0L, DateTimeKind.Utc);
        public static DateTime MinUtcDateTime => _minUtcDateTime;
        
        private static TimeSpan _serverClientTimeDiff;
        private static DateTime _serverTime;

        private static DateTime ServerTime
        {
            get { return _serverTime; }

            set
            {
                _serverTime = value;
                _serverClientTimeDiff = DateTime.UtcNow.Subtract(value);

                Log.Info("Server Time is set to " + ServerTimeNow.ToString("G", CultureInfo.InvariantCulture) + " Time Difference is " + _serverClientTimeDiff.TotalSeconds);
            }
        }

        public static DateTime ServerTimeNow
        {
            get
            {
                if (_fetchServerTimeStatus == ServerTimeStatus.Initialized)
                {
                    return DateTime.UtcNow.Subtract(_serverClientTimeDiff);
                }

                return MinUtcDateTime;
            }
        }

        public static async void FetchServerDateTimeAsync(int timeOut = 10 * 1000, bool forceFetch = false)
        {
            try
            {
                if (FetchServerTimeStatus == ServerTimeStatus.Initializing)
                    return;

                if (!forceFetch)
                {
                    if (FetchServerTimeStatus == ServerTimeStatus.Initialized)
                        return;
                }

                FetchServerTimeStatus = ServerTimeStatus.Initializing;

                using var unityWebRequest = UnityWebRequest.Get("https://gm.gamepantheon.com/getServerTime");
                unityWebRequest.timeout = timeOut;
                unityWebRequest.SetRequestHeader("Content-Type", "application/json");

                await unityWebRequest.SendWebRequest();
                if (unityWebRequest.result == UnityWebRequest.Result.Success)
                {
                    var jsonObject = JsonUtility.FromJson<ServerDateTime>(unityWebRequest.downloadHandler.text);
                    if (long.TryParse(jsonObject.timestamp, out var unixTimeStampInMillieSeconds))
                    {
                        var dateTimeOffset = DateTimeOffset.FromUnixTimeMilliseconds(unixTimeStampInMillieSeconds);
                        ServerTime = dateTimeOffset.DateTime;
                        FetchServerTimeStatus = ServerTimeStatus.Initialized;
                    }
                    else
                    {
                        Log.Error("Failed to parse unix millie seconds while trying to get date time from server!");
                        FetchServerTimeStatus = ServerTimeStatus.Failed;
                    }
                }
                else
                {
                    Log.Error($"Failed to connect server time endpoint {unityWebRequest.error}");
                    FetchServerTimeStatus = ServerTimeStatus.Failed;
                }
            }
            catch (Exception e)
            {
                Log.Exception($"Error while fetching time from server. Reason: {e.Message}");
            }
        }
        #endregion

        private const double _clientApiVersion = 1.3;
#if ROC_LOCAL
		public static readonly string RocGmBaseUrl = "https://test.zeus.panteon.games";
		public static readonly string RocGmTrackerUrl = "https://test.zeus.panteon.games/tracker";
#else
        //TODO Ahmet
        public static readonly string RocGmBaseUrl = "https://gm.gamepantheon.com";
        public static readonly string RocGmTrackerUrl = "https://gm.gamepantheon.com/tracker";
        // public static readonly string RocGmBaseUrl = "https://test.zeus.panteon.games";
        // public static readonly string RocGmTrackerUrl = "http://192.168.168.192:8081/tracker";
#endif


        public static int SimulatePlatform = 0;
        private const int RetryIntervalInSeconds = 10;

        private const string GameSettingsKey = "gameSettings";
        private const string ServerTimeKey = "serverTime";
        private const string PlayerDataKey = "remotePlayerData";
        public const string PlayerDataIsDirtyKey = "gmPlayerDataIsDirty";
        public const string PlayerDataContentKey = "data";
        public const string PlayerDataDiffCountKey = "remotePlayerDataDiffCount";
        public const string MaintenanceKey = "isMaintenance";

        public static string EmptyRemotePlayerData = "{}";
        public static string RemotePlayerData = EmptyRemotePlayerData;

        public static bool InitializeAutomatically;
        public static bool DisableRemotePlayerData;

        public static bool IsRemotePlayerDataEmpty => RemotePlayerData == EmptyRemotePlayerData || RemotePlayerData == null || RemotePlayerData == string.Empty;

        protected static readonly RocLog Log = new RocLog(nameof(RocGm), DebugLevels.Debug);

        public RocGm(RocGmGameSettings gameSettings)
        {
            PlayerId = 0;
            _parameters = null;
            _status = InitializationStatus.NotStarted;
            CoroutineController.Reset();

            GameSettings = gameSettings;
            PlayerProperties = new RocGmPlayerProperties();

            Reporting = new RocGmReporting();
            LiveTune = new RocGmLiveTune();
            PlayerProgress = new RocGmPlayerProgress(gameSettings.PersistentProgressEnabled);
        }

#if ROC_GM_ADJUST
		public static RefSourceConverterBase RefSourceConverter = new RefSourceConverterBase();
#endif
        public static RocGmConfig Config { get; private set; }

		#region Initialize
        public static void Initialize(RocGmConfig config)
        {
            if (IsInitialized)
            {
                Log.Warning("Already initialized!", "Initialize");
                return;
            }

            Config = config;

            if (Config.GameId == 0)
            {
                Log.Error("Please change default gameId to a valid value!");
                return;
            }

            if (!InternetConnectionController.HasConnection)
            {
                Status = InitializationStatus.WaitingToConnectInternet;
            }

            InternetConnectionController.DoAfterConnectionAvailable(() =>
            {
                Log.Debug("Initialization is started");
                Status = InitializationStatus.Initializing;

                CoroutineController.StartCoroutine(InitializeRoutine());
            });
        }

        public static void SetConfig(RocGmConfig config)
        {
            Config = config;
        }
        
        public static void InitializeWithPackageInitializationData(Dictionary<string, object> initializationResultParameters)
        {
            if (IsInitialized)
            {
                Log.Warning("Already initialized!", "InitializeFromResult");
                return;
            }

            if (Config.GameId == 0)
            {
                Log.Error("Please change default gameId to a valid value!");
                return;
            }

            if (!InternetConnectionController.HasConnection)
            {
                Status = InitializationStatus.WaitingToConnectInternet;
            }

            _parameters = initializationResultParameters;

            InternetConnectionController.DoAfterConnectionAvailable(() =>
            {
                Log.Debug("Initialization is started");
                Status = InitializationStatus.Initializing;
                
                
                bool advertisingIdWillBeUpdated = !string.IsNullOrEmpty(Config.AdvertisingId);
                
                InitializeUsingParameters(advertisingIdWillBeUpdated);
            });
        }

        private static IEnumerator InitializeRoutine()
        {
            if (IsInitialized)
            {
                Log.Warning("Already initialized!", "InitializeRoutine");
                yield break;
            }

            if (Application.platform == RuntimePlatform.IPhonePlayer)
            {
                yield return new WaitUntilWithTimeout(5, () => Config.IsAdvertisingIdReceived);
            }

#if ROC_GM_ADJUST
			Log.Debug(string.Format(
				"Initialize: \nGameId:{0}\nSecretKey:{1}\nPlatform:{2}\nDeviceId:{3}\nAdvertisingId:{4}\nAdjustId:{5}\nRefSource:{6}\nRefCampaign:{7}\nRefAdGroup:{8}\nRefCreative:{9}" +
				" \nUserAge:{10}\nLang:{11}\nDeviceModel:{12}\nDeviceOs:{13}",
				Config.GameId, Config.SecretKey, Config.PlatformId, Config.DeviceId, Config.AdvertisingId, Config.AdjustId,
				Config.GetNonEmptyRefSource(), Config.RefCampaign, Config.RefAdGroup, Config.RefCreative, Config.UserAge, Application.systemLanguage.ToString(),
				Config.DeviceModel, Config.DeviceOs));
#else
            Log.Debug(string.Format(
                "Initialize: \nGameId:{0}\nSecretKey:{1}\nPlatform:{2}\nDeviceId:{3}\nAdvertisingId:{4}\nUserAge:{5}\nLang:{6}\nDeviceModel:{7}\nDeviceOs:{8}",
                Config.GameId, Config.SecretKey, Config.PlatformId, Config.DeviceId, Config.AdvertisingId, Config.UserAge, Application.systemLanguage.ToString(),
                Config.DeviceModel, Config.DeviceOs));
#endif


            WWWForm request = CreateRequest();

            request.AddField("platform", Config.PlatformId);
            request.AddField("environment", Config.Environment);
            request.AddField("deviceId", Config.DeviceId);
            request.AddField("userAge", Config.UserAge);
            request.AddField("lang", Config.Language);
            request.AddField("deviceModel", Config.DeviceModel);
            request.AddField("deviceOs", Config.DeviceOs);
            request.AddField("clientVersion", Config.ClientVersion.ToString(CultureInfo.InvariantCulture));
            request.AddField("locale", Config.Locale);

            #if ROC_DEBUG_MODE
                request.AddField("isTestDevice", "true");
                Config.IsTestDevice = true;
            #endif

            bool advertisingIdWillBeUpdated = false;

            if (!string.IsNullOrEmpty(Config.AdvertisingId))
            {
                advertisingIdWillBeUpdated = true;
                request.AddField("advertiserId", Config.AdvertisingId);
            }

#if ROC_GM_ADJUST
			if (!string.IsNullOrEmpty(Config.AdjustId))
			{
				request.AddField("adjustId", Config.AdjustId);
			}

			if (!string.IsNullOrEmpty(Config.RefSource))
			{
                string refSource = Config.GetNonEmptyRefSource();
				RefSource convertedRefSource = RefSourceConverter.Convert(refSource);

				if (convertedRefSource == RefSource.None)
				{
					request.AddField("refSource", refSource);
				}
				else
				{
					request.AddField("refSource", (int)convertedRefSource);
				}
			}			

			if (!string.IsNullOrEmpty(Config.RefCampaign))
			{
				request.AddField("refCampaign", Config.RefCampaign);
			}

			if (!string.IsNullOrEmpty(Config.RefAdGroup))
			{
				request.AddField("refAdGroup", Config.RefAdGroup);
			}

			if (!string.IsNullOrEmpty(Config.RefCreative))
			{
				request.AddField("refCreative", Config.RefCreative);
			}
#elif !ROC_GM_APPSFLYER
            request.AddField("refSource", "NotIntegrated");
			request.AddField("adjustId", "NotIntegrated");
#endif

            UnityWebRequest www = UnityWebRequest.Post($"{RocGmTrackerUrl}/initialize", request);

            www.chunkedTransfer = false;
            www.useHttpContinue = false;

            Log.Debug("Sending INITIALIZATION request to server: " + RocGmTrackerUrl);

            //Only do this operation once!
            if (!_failedOnce)
            {
                www.timeout = 5;
            }
            else
            {
                Status = InitializationStatus.RetryingToSolveConnectionProblems;
                www.timeout = 10;
            }

            yield return www.SendWebRequest();

            Log.Debug("Received initialization response.");

            if (www.responseCode != 200 || !RocGmHelper.TryGetServerResult(www.downloadHandler.text, out _parameters))
            {
                Log.Warning("Initialization FAILED!" +
                    $" responseCode: {www.responseCode}" +
                    $" error: {www.error}" +
                    $" text: {www.downloadHandler.text}");
                OnInitializationFailed();
                yield break;
            }

            InitializeUsingParameters(advertisingIdWillBeUpdated);

            www.Dispose();
        }

        private static void InitializeUsingParameters(bool advertisingIdWillBeUpdated)
        {
            PlayerId = (long)_parameters["playerId"];

            if (_parameters.ContainsKey(MaintenanceKey))
                IsMaintenance = (bool)_parameters[MaintenanceKey];

            if (_parameters.ContainsKey("experimentId"))
            {
                Config.ExperimentId = Convert.ToInt32(_parameters["experimentId"]);
                Config.ExperimentVariant = Convert.ToInt32(_parameters["experimentVariant"]);
            }

            if (_parameters.TryGetValue("token", out var secretKey))
            {
                Config.UpdateSecretKeyAndGenerateSecurityKey(secretKey.ToString());
                Log.Debug("Secret key updated  and security generated!");
            }

            if (_parameters.TryGetValue("isTestDevice", out var isTestDeviceObject))
            {
                Config.IsTestDevice = Convert.ToBoolean(isTestDeviceObject);
            }

            // For space time complexity.

            if (_parameters.TryGetValue("playerProperties", out var playerProperties))
            {
                var playerPropertiesString = playerProperties.ToString();
                if (Config.IsEncryptedMessagingActive)
                {
                    playerPropertiesString = RocGmSecurity.Decrypt(playerPropertiesString, Config.SecurityKey);
                }
                var playerPropertiesObject = JObject.Parse(playerPropertiesString);
                if (playerPropertiesObject["available"] is JArray availablePlayerPropertiesJArray &&
                    playerPropertiesObject["current"] is JArray currentPlayerPropertiesJArray)
                {
                    var availablePlayerProperties = availablePlayerPropertiesJArray.ToObject<List<RocGmPlayerProperty>>();
                    var currentPlayerProperties = currentPlayerPropertiesJArray.ToObject<List<RocGmPlayerProperty>>();
                    PlayerProperties.CopyFrom(availablePlayerProperties, currentPlayerProperties);
                }
            }

            if (advertisingIdWillBeUpdated)
                Config.IsAdvertisingIdUpdated = true;

            GameSettings.CopyFrom(JsonUtility.FromJson<RocGmGameSettings>(_parameters[GameSettingsKey].ToString()));
            GameSettings.OverrideRemotePlayerData(!DisableRemotePlayerData);

            // Synchronize with server time
            if (_parameters.ContainsKey(ServerTimeKey))
            {
                var serverTime = DateTimeExtensions.FromUnixTimestamp(Convert.ToInt64(_parameters[ServerTimeKey]));

                ServerTime = serverTime;
                FetchServerTimeStatus = ServerTimeStatus.Initialized;
            }

            if (GameSettings.RemotePlayerDataEnabled)
            {
                if (_parameters.TryGetValue(PlayerDataKey, out var remotePlayerData))
                {
                    var remotePlayerDataJObject = JObject.Parse(RocGmSecurity.Decrypt(remotePlayerData.ToString(), Config.SecurityKey));
                    var remotePlayerDataString = remotePlayerDataJObject[PlayerDataContentKey]?.Value<string>();
                    var remotePlayerDataDiffCountString = remotePlayerDataJObject[PlayerDataDiffCountKey]?.Value<string>();
                    var isPlayerDataDirty = remotePlayerDataJObject[PlayerDataIsDirtyKey]?.Value<bool>();

                    if (remotePlayerDataString == null || remotePlayerDataDiffCountString == null)
                        return;

                    RemotePlayerData = remotePlayerDataString;
                    ulong.TryParse(remotePlayerDataDiffCountString, out var remoteDataDiffCount);

                    if (isPlayerDataDirty == true)
                    {
                        //Clear Session Data if player data is dirty
                        var unsentSessionData = RocGmSessionReporting.GetSentSessionDataContainer();
                        if (unsentSessionData.SentImpressionDataList != null)
                        {
                            for (var index = 0; index < unsentSessionData.SentImpressionDataList.Count; index++)
                            {
                                SentSessionData sentSessionData = unsentSessionData.SentImpressionDataList[index];
                                sentSessionData.RemotePlayerDataDiff = "";
                                unsentSessionData.SentImpressionDataList[index] = sentSessionData;
                            }

                            RocGmSessionReporting.SentSessionData = JsonUtility.ToJson(unsentSessionData);
                        }
                    }

                    if (isPlayerDataDirty == true || RocGmSessionReporting.RemotePlayerDataDiffCount == 0)
                    {
                        RocGmSessionReporting.RemotePlayerDataDiffCount = remoteDataDiffCount;
                    }
                }


            }
            else
            {
                RocGmSessionReporting.RemotePlayerDataDiffCount = 0;
            }

            Log.Debug("Initialize event invoked!");

            Initialized?.Invoke(_parameters, GameSettings);
            Status = InitializationStatus.Initialized;

            Log.Debug("Initialization to " +
#if ROC_LOCAL
			    "local"
#else
                "GM"
#endif
                + " server completed! PlayerId: " + PlayerId, "InitializeRoutine");
        }

        #endregion

		#region Helpers
        public static void OpenStorePage()
        {
            string storeUrl;
            switch (Application.platform)
            {
                case RuntimePlatform.Android:
                    storeUrl = "https://play.google.com/store/apps/details?id=" + GameSettings.AppStoreId;
                    break;
                case RuntimePlatform.IPhonePlayer:
                    storeUrl = "https://itunes.apple.com/app/id" + GameSettings.AppStoreId;
                    break;
                default:
                    storeUrl = "https://itunes.apple.com/app/id" + GameSettings.AppStoreId;
                    break;
            }

            Application.OpenURL(storeUrl);
        }

#if ROC_GM_ADJUST
		public static void SetRefSourceConverter(RefSourceConverterBase newRefSourceConverter)
		{
			RefSourceConverter = newRefSourceConverter;
		}
#endif
        public static WWWForm CreateRequest()
        {
            WWWForm request = new WWWForm();

            request.AddField("clientApiVersion", _clientApiVersion.ToString(CultureInfo.InvariantCulture));
            request.AddField("clientVersion", Config.ClientVersion.ToString(CultureInfo.InvariantCulture));
            request.AddField("secretKey", Config.SecretKey);
            request.AddField("gameId", Config.GameId);
            request.AddField("platform", Config.PlatformId);

            return request;
        }

        private static void OnInitializationFailed()
        {
            _failedOnce = true;

            CoroutineController.DoAfterGivenTime(RetryIntervalInSeconds, () =>
            {
                if (InitializeAutomatically)
                {
                    Initialize(Config);
                }
                else
                {
                    InitializeWithPackageInitializationData(_parameters);
                }
            });

            Status = InitializationStatus.Failed;
        }

        /// <summary>
        /// Calls given action if/when remote player id comes from reporting server.
        /// </summary>
        /// <param name="action"></param>
        public static void DoAfterInitialized(Action<Dictionary<string, object>, RocGmGameSettings> action)
        {
            if (IsInitialized)
            {
                action(_parameters, GameSettings);
            }
            else
            {
                Action<Dictionary<string, object>, RocGmGameSettings> eventHandler = null;
                eventHandler = delegate
                {
                    action(_parameters, GameSettings);
                    Initialized -= eventHandler;
                };

                Initialized += eventHandler;
            }
        }

        public static void DoAfterInitialized(Action action)
        {
            if (IsInitialized)
            {
                action();
            }
            else
            {
                Action<Dictionary<string, object>, RocGmGameSettings> eventHandler = null;
                eventHandler = delegate
                {
                    action();
                    Initialized -= eventHandler;
                };

                Initialized += eventHandler;
            }
        }
        #endregion

		#region PlayerProperty
        internal static async void SendPlayerProperty(int id, string value)
        {
            if (!IsInitialized)
            {
                Log.Warning("Can not send player property, wait for GM to be initialized!");
                return;
            }

            var webRequest = CreateRequest();
            webRequest.AddField("playerId", RocGm.PlayerId.ToString());

            var propertyDic = new Dictionary<string, Object>();
            propertyDic.Add("propertyId", id);
            propertyDic.Add("propertyValue", value);

            var propertyJson = JObject.FromObject(propertyDic).ToString();

            if (Config.IsEncryptedMessagingActive)
                propertyJson = RocGmSecurity.Encrypt(propertyJson, Config.SecurityKey);

            webRequest.AddField("property", propertyJson);

            var unityWebRequest = UnityWebRequest.Post($"{RocGmTrackerUrl}/setPlayerProperty", webRequest);
            await unityWebRequest.SendWebRequest();

            if (unityWebRequest.result == UnityWebRequest.Result.Success)
            {
                SentPlayerPropertyToGameMaster?.Invoke();
                Log.Debug("Sent player property to GM.");
            }
            else
            {
                Log.Warning($"Failed to send player property to the GM, the reason was {unityWebRequest.error}");
            }
        }
		#endregion

		#region Device
        public static async void SendDeviceProperties(string deviceName, string deviceID)
        {
            throw new Exception("Do not use this function, waiting for end-point url and implementation is not ready yet!");

            // var webRequest = CreateRequest();
            // webRequest.AddField("DeviceName", deviceName);
            // webRequest.AddField("DeviceID",deviceID);
            //
            // var unityWebRequest = UnityWebRequest.Post("EndpointGoesHere", webRequest);
            // await unityWebRequest.SendWebRequest();
            //
            // if (unityWebRequest.result == UnityWebRequest.Result.Success)
            // {
            // 	Log.Debug("Sent device properties to GM.");
            // }
            // else
            // {
            // 	Log.Warning($"Failed to send device properties to the GM, the reason was {unityWebRequest.error}");
            // }
        }
		#endregion

        #region Simulation
#if UNITY_EDITOR
        public static void FailInitialize(RocGmConfig config)
        {
            Config = config;

            CoroutineController.DoAfterGivenTime(2f, () =>
            {
                Log.Error("SIMULATE INITIALIZE FAIL");
                OnInitializationFailed();
            });
        }
#endif
        #endregion

        public static async Task SendDeletePlayerDataRequest()
        {
            if (!GameSettings.RemotePlayerDataEnabled)
                return;

            var webRequest = CreateRequest();
            webRequest.AddField("playerId", RocGm.PlayerId.ToString());

            var unityWebRequest = UnityWebRequest.Post($"{RocGmTrackerUrl}/deleteRemotePlayerData", webRequest);
            await unityWebRequest.SendWebRequest();

            if (unityWebRequest.result == UnityWebRequest.Result.Success)
            {
                RemotePlayerData = EmptyRemotePlayerData;
                Log.Debug("Delete player data request is success!");
            }
            else
            {
                Log.Warning($"Failed to send delete player data request to the GM, the reason was {unityWebRequest.error}");
            }
        }
    }
}
