﻿using System;
using System.Globalization;
using System.Text;
using System.Threading;
using JetBrains.Annotations;
using RocketCoroutine;
using RocketLog;
using RocketUtils;
using UnityEngine;

namespace RocketGM
{
    public class RocGmConfig
    {
        public static event Action AdvertisingIdReceived;

        public int GameId { get; private set; }
        public string SecretKey { get; private set; }
        public byte[] SecurityKey { get; private set; }
        public bool IsEncryptedMessagingActive => SecurityKey != null;

        public int PlatformId
        {
            get
            {
                switch (Application.platform)
                {
                    case RuntimePlatform.Android:
                        return 2;
                    case RuntimePlatform.IPhonePlayer:
                        return 1;
                    default:
                        return RocGm.SimulatePlatform;
                }
            }
        }

        public byte Environment
        {
            get
            {
#if ROC_DEBUG_MODE
                return 1; // Development
#else
                return 3; // Release
#endif

            }
        }

        public string AdvertisingId
        {
            get { return PlayerPrefs.GetString(AdvertisingIdKey, null); }
            set { PlayerPrefs.SetString(AdvertisingIdKey, value); }
        }
        public bool IsAdvertisingIdReceived;

        public string DeviceId { get; private set; }
        public string DeviceModel { get; private set; }
        public string DeviceOs { get; private set; }
        public string Language { get; private set; }

#if ROC_GM_ADJUST
        public string AdjustId
        {
            get { return PlayerPrefs.GetString(AdjustIdKey, null); }
            set { PlayerPrefs.SetString(AdjustIdKey, value); }
        }
#endif

#if ROC_GM_APPSFLYER
        public string ConversionData
        {
            get { return PlayerPrefs.GetString(ConversionDataKey, null); }
            set { PlayerPrefs.SetString(ConversionDataKey, value); }
        }
#endif

#if ROC_GM_ADJUST || ROC_GM_APPSFLYER
        public string RefSource
        {
            get { return PlayerPrefs.GetString(RefSourceKey, null); }
            set { PlayerPrefs.SetString(RefSourceKey, value); }
        }

        public string RefCampaign
        {
            get { return PlayerPrefs.GetString(RefCampaignKey, null); }
            set { PlayerPrefs.SetString(RefCampaignKey, value); }
        }

        public string RefAdGroup
        {
            get { return PlayerPrefs.GetString(RefAdGroupKey, null); }
            set { PlayerPrefs.SetString(RefAdGroupKey, value); }
        }

        public string RefCreative
        {
            get { return PlayerPrefs.GetString(RefCreativeKey, null); }
            set { PlayerPrefs.SetString(RefCreativeKey, value); }
        }
#endif

        public string PushToken
        {
            get { return PlayerPrefs.GetString(PushTokenKey, null); }
            set { PlayerPrefs.SetString(PushTokenKey, value); }
        }

        public float ClientVersion
        {
            get { return float.Parse(Application.version, CultureInfo.InvariantCulture); }
        }
        
        public string Locale
        {
            get { return CultureInfo.CurrentUICulture.ToString(); }
        }

        public int UserAge
        {
            get
            {
                var userAgeSpan = DateTime.UtcNow - InitialLaunchTime;
                return (int)userAgeSpan.TotalDays;
            }
        }

        private DateTime InitialLaunchTime
        {
            get
            {
                if (!PlayerPrefs.HasKey(InitialLaunchTimeKey)) return DateTime.UtcNow;
                string initialLaunchTimeValue = PlayerPrefs.GetString(InitialLaunchTimeKey);
                DateTime targetDate;
                bool isValid = DateTime.TryParse(initialLaunchTimeValue, CultureInfo.InvariantCulture, DateTimeStyles.None, out targetDate);
                if (!isValid) return DateTime.UtcNow;

                return targetDate;
            }
            set
            {
                CultureInfo currentCulture = Thread.CurrentThread.CurrentCulture;
                Thread.CurrentThread.CurrentCulture = new CultureInfo("en-GB");

                PlayerPrefs.SetString(InitialLaunchTimeKey, value.ToUniversalTime().ToLongDateString());

                Thread.CurrentThread.CurrentCulture = currentCulture;
            }
        }

        public DateTime LastActiveAt
        {
            get
            {
                if (!PlayerPrefs.HasKey(LastActiveAtKey)) return DateTime.UtcNow;
                string lastActiveAtValue = PlayerPrefs.GetString(LastActiveAtKey);
                DateTime targetDate;
                bool isValid = DateTime.TryParse(lastActiveAtValue, CultureInfo.InvariantCulture, DateTimeStyles.None, out targetDate);
                if (!isValid) return DateTime.UtcNow;

                return targetDate;
            }
            set
            {
                CultureInfo currentCulture = Thread.CurrentThread.CurrentCulture;
                Thread.CurrentThread.CurrentCulture = new CultureInfo("en-GB");

                PlayerPrefs.SetString(LastActiveAtKey, value.ToLongDateString());

                Thread.CurrentThread.CurrentCulture = currentCulture;
            }
        }

        public bool IsAttributionUpdated
        {
            get { return PlayerPrefs.GetInt(IsAttributionUpdatedKey, 0) == 1; }
            set { PlayerPrefs.SetInt(IsAttributionUpdatedKey, value ? 1 : 0); }
        }

        public bool IsPushTokenSent
        {
            get { return PlayerPrefs.GetInt(IsPushTokenSentKey, 0) == 1; }
            set { PlayerPrefs.SetInt(IsPushTokenSentKey, value ? 1 : 0); }
        }

        public string SavedDeviceId
        {
            get { return PlayerPrefs.GetString(DeviceIdKey, ""); }
            set { PlayerPrefs.SetString(DeviceIdKey, value); }
        }

        public bool IsAdvertisingIdUpdated
        {
            get { return PlayerPrefs.GetInt(AdvertisingIdUpdatedKey, 0) == 1; }
            set { PlayerPrefs.SetInt(AdvertisingIdUpdatedKey, value ? 1 : 0); }
        }
        
        public bool IsTestDevice
        {
            get { return PlayerPrefs.GetInt(TestDeviceKey, 0) == 1; }
            set { PlayerPrefs.SetInt(TestDeviceKey, value ? 1 : 0); }
        }

        public bool IsFirstSession { get; private set; }
        public int ExperimentId { get; set; }
        public int ExperimentVariant { get; set; }

#if ROC_GM_ADJUST
		private const string AdjustIdKey = "gmAdjustId";
#endif

#if ROC_GM_APPSFLYER
        private const string ConversionDataKey = "gmConversionData";
#endif

#if ROC_GM_ADJUST || ROC_GM_APPSFLYER
        private const string RefSourceKey = "gmRefSource";
        private const string RefCampaignKey = "gmRefCampaign";
        private const string RefAdGroupKey = "gmRefAdGroup";
        private const string RefCreativeKey = "gmRefCreative";
#endif
        private const string InitialLaunchTimeKey = "gmInitialLaunchTime";
        private const string IsFirstSessionKey = "gmIsFirstTime";
        private const string LastActiveAtKey = "gmLastActiveAt";
        private const string IsAttributionUpdatedKey = "gmIsAttributionUpdated";
        private const string PushTokenKey = "gmPushToken";
        private const string IsPushTokenSentKey = "gmIsPushTokenSent";
        private const string DeviceIdKey = "gmDeviceId";
        private const string AdvertisingIdUpdatedKey = "gmAdvertisingIdUpdated";
        private const string AdvertisingIdKey = "gmAdvertisingId";
        private const string TestDeviceKey = "testDevice";

        //Dont forget to change this when server side key changes
        private const string ConstantSecretKey = "h=J9B5w$";

        protected RocLog Log = new RocLog("RocGmConfig", DebugLevels.Debug);

        public RocGmConfig(int gameId, string secretKey)
        {
            GameId = gameId;
            SecretKey = secretKey;

            DeviceId = SavedDeviceId;

            if (string.IsNullOrEmpty(DeviceId))
            {
                DeviceId = DeviceUtility.GetDeviceUniqueId();
                SavedDeviceId = DeviceId;
            }

            Language = Application.systemLanguage.ToString();

            DeviceModel = DeviceUtility.GetDeviceModel();
            DeviceOs = DeviceUtility.GetDeviceOs();

            if (PlayerPrefs.HasKey(IsFirstSessionKey))
            {
                IsFirstSession = false;
            }
            else
            {
                IsFirstSession = true;
                PlayerPrefs.SetInt(IsFirstSessionKey, 0);
                InitialLaunchTime = DateTime.UtcNow;
            }

            GetAdvertisingId();
        }

        private void GetAdvertisingId()
        {
#if UNITY_EDITOR
            AdvertisingId = "UnityEditor";
#elif PLATFORM_STANDALONE_WIN
			AdvertisingId = "Standalone";
#else
#if UNITY_ANDROID
            TryGetAdvertisingIdForAndroid();
#else
            Application.RequestAdvertisingIdentifierAsync(OnAdvertisingIdReceived);
#endif // UNITY_ANDROID
#endif
        }

        private string TryGetAdvertisingIdForAndroid()
        {
            var advertisingId = string.Empty;
            try
            {
                var up = new AndroidJavaClass("com.unity3d.player.UnityPlayer");
                var currentActivity = up.GetStatic<AndroidJavaObject>("currentActivity");
                var client = new AndroidJavaClass("com.google.android.gms.ads.identifier.AdvertisingIdClient");
                var adInfo = client.CallStatic<AndroidJavaObject>("getAdvertisingIdInfo", currentActivity);

                advertisingId = adInfo.Call<string>("getId").ToString();
                OnAdvertisingIdReceived(advertisingId, true, string.Empty);
            }
            catch (Exception exception)
            {
                Log.Warning($"Failed to fetch Advertising Id for Android. The exception was: {exception.Message}");
            }

            return advertisingId;
        }

        private void OnAdvertisingIdReceived(string advertisingId, bool trackingEnabled, string error)
        {
            Log.Debug(string.Format("Advertising Id Received. Id: {0}, TrackingEnabled: {1}", advertisingId, trackingEnabled));

            if (advertisingId == "00000000-0000-0000-0000-000000000000")
            {
                IsAdvertisingIdReceived = true;
                AdvertisingId = null;
                return;
            }

            IsAdvertisingIdReceived = true;
            AdvertisingId = advertisingId;

            if (AdvertisingIdReceived != null) AdvertisingIdReceived();
        }

        public string GetNonEmptyRefSource()
        {
#if ROC_GM_ADJUST || ROC_GM_APPSFLYER
            return string.IsNullOrEmpty(RefSource) ? "Organic" : RefSource;
#else
            return "NotIntegrated";
#endif
        }

        public void UpdateSecretKeyAndGenerateSecurityKey(string secretKey)
        {
            SecretKey = secretKey;

            var sb = new StringBuilder();
            sb.Append(ConstantSecretKey);
            sb.Append(secretKey.Substring(0, 16 - ConstantSecretKey.Length)); // AES 128 key must be 16 digits.

            SecurityKey = Encoding.UTF8.GetBytes(sb.ToString());
        }

    }
}
