﻿using System;
using System.Globalization;
using System.Net;
using System.Net.Security;
using System.Security.Cryptography.X509Certificates;
using RocketCoroutine;
using RocketGM.RemoteSettings;
using RocketLog;
using RocketUtils;
using UnityEngine;
using UnityEngine.Serialization;

namespace RocketGM
{
    [Serializable]
    public class GameSettings
    {
        public string AppStoreId = "";
#if UNITY_EDITOR
        public bool SimulateLateInitialize;
#endif
        public bool PersistentProgressEnabled;
#if ROC_SUBSCRIPTION
        public bool SubscriptionEnabled;
#else
        [NonSerialized]
        public bool SubscriptionEnabled = false;
#endif
        [NonSerialized]
        public bool LegacyMarketingEnabled = false;

        public bool RemotePlayerDataEnabled = false;

        public bool InitializeAutomatically = true;

        public bool DisableRemotePlayerData = false;
    }

    public class RocGmControllerBase : SingletonBehaviour<RocGmControllerBase>
    {
        public static event Action Initialized;
        public static event Action<bool> PauseStateChanged;
        public static event Action<bool> FocusStateChanged;
        public static event Action ApplicationQuit;
        public static event Action UpdateRequired;
        public static event Action BadConnectionOnInitialize;
        public static event Action<int, int> ResolutionChanged;
        public static event Action ForcePlayerToCloseGame;

        [SerializeField]
        [Tooltip("Rocket GM GameId")]
#pragma warning disable 0649
        private int _gameId;

        [SerializeField]
        [Tooltip("Rocket GM Secret Key")]
#pragma warning disable 0649
        private string _secretKey;

        [SerializeField]
        [Tooltip("Rocket GM Game Settings")]
#pragma warning disable 0649
        private GameSettings _gameSettings;

        public static bool IsInitialized
        {
            get { return _isInitialized; }
            set
            {
                if (_isInitialized == value) return;

                _isInitialized = value;

                if (Initialized != null) Initialized();
            }
        }
        private static bool _isInitialized;

        public static bool IsCustomBuildNumberUsageForced
        {
            get => PlayerPrefs.GetInt("IsCustomBuildNumberUsageForced", 0) == 1;
            set => PlayerPrefs.SetInt("IsCustomBuildNumberUsageForced", value ? 1 : 0);
        }

        public static float CustomBuildNumber
        {
            get => PlayerPrefs.GetFloat("CustomBuildNumber", 0.0f);
            set => PlayerPrefs.SetFloat("CustomBuildNumber", value);
        }

        private bool _firstFrame;

        public readonly RemoteBool IdfaAuthenticationOnInitialize = new RemoteBool("idfaAuthenticationOnInitialize", true);

        protected RocLog Log = new RocLog("RocGmControllerBase", DebugLevels.Debug);

        protected override void OnAwake()
        {
            // Do not destroy base object.
            DontDestroyOnLoad(gameObject);

            _firstFrame = true;

            Log = new RocLog(GetType().Name, DebugLevels.Debug);

            CoroutineController.DoAfterFixedUpdate(() =>
            {
                _firstFrame = false;
                Log.Debug("First frame finished!");
            });

            RocGmSessionReporting.PlayerDataIsDirty += OnPlayerDataIsDirty;
            RocGmSessionReporting.MaintenanceActivated += OnMaintenanceActivated;
            InternetConnectionController.Initialize();

            // SSL Certificate Validation
            ServicePointManager.ServerCertificateValidationCallback = RemoteCertificateValidationCallback;

            string appStoreId = "";

            switch (Application.platform)
            {
                case RuntimePlatform.Android:
                case RuntimePlatform.WindowsEditor:
                    appStoreId = Application.identifier;
                    break;
                case RuntimePlatform.IPhonePlayer:
                    appStoreId = _gameSettings.AppStoreId;
                    break;
            }

            RocGmGameSettings gmGameSettings = new RocGmGameSettings(appStoreId, _gameSettings.LegacyMarketingEnabled, _gameSettings.SubscriptionEnabled, _gameSettings.PersistentProgressEnabled, _gameSettings.RemotePlayerDataEnabled);

            RocGm gm = new RocGm(gmGameSettings);
            RocGm.LiveTune.QualitySettingsApplied += OnQualitySettingsApplied;
        }
        
        private void OnMaintenanceActivated()
        {
            ForcePlayerToCloseGame?.Invoke();
        }
        
        private void OnPlayerDataIsDirty()
        {
            ForcePlayerToCloseGame?.Invoke();
        }

        protected virtual void OnDestroy()
        {
            if (RocGm.LiveTune != null)
                RocGm.LiveTune.QualitySettingsApplied -= OnQualitySettingsApplied;
        }

        protected virtual void Start()
        {
            Initialize();
        }

        private bool TryAuthorizeClient()
        {
            if (IdfaAuthenticationOnInitialize)
            {
                RocGmIDFAController.TryToInitiateAuthorization(true, true);
            }

            bool? isUpdateRequired = IsUpdateRequired();
            if (isUpdateRequired.HasValue && isUpdateRequired.Value)
            {
                Log.Warning("Min Client Version requirement does NOT satisfied!");
                if (UpdateRequired != null)  UpdateRequired.Invoke();
                
                return false;
            }

            return true;
        }

        private void Initialize()
        {
            Log.Debug("Initializing", "RocGm");

            RocGmConfig config = new RocGmConfig(_gameId, _secretKey);

            if (RocGm.GameSettings.RemotePlayerDataEnabled)
            {
                Log.Debug("Remote Data Sync is enabled! Waiting for gm initialization!", "RocGm");
                RocGm.DoAfterInitialized(() =>
                {
                    Log.Debug($"RocGmControllerBase DoAfterInitialized", nameof(Start));

                    if(!TryAuthorizeClient())
                        return;
                    IsInitialized = true;
                    Log.Debug("Initialized", "RocGm");
                });
                RocGm.StatusChanged += OnRocGmStatusChanged;
            }
            else
            {
                RocGm.RemoteSettings.DoAfterInitialized(() =>
                {
                    Log.Debug($"RocGmControllerBase DoAfterInitialized", nameof(Start));

                    if(!TryAuthorizeClient())
                        return;;
                    IsInitialized = true;
                });
            }

#if UNITY_EDITOR
            if (_gameSettings.SimulateLateInitialize)
            {
                RocGm.FailInitialize(config);
                return;
            }
#endif // UNITY_EDITOR

            RocGm.SetConfig(config);
            RocGm.InitializeAutomatically = _gameSettings.InitializeAutomatically;
            
            if (_gameSettings.InitializeAutomatically)
                RocGm.Initialize(config);
        }


        private void OnRocGmStatusChanged(RocGm.InitializationStatus initializationStatus)
        {
            if (initializationStatus == RocGm.InitializationStatus.Initialized)
            {
                RocGm.StatusChanged -= OnRocGmStatusChanged;
                return;
            }

            if (initializationStatus == RocGm.InitializationStatus.Failed || initializationStatus == RocGm.InitializationStatus.WaitingToConnectInternet)
            {
                Log.Debug($"BadConnectionOnInitialize. Status:{initializationStatus}", nameof(OnRocGmStatusChanged));
                BadConnectionOnInitialize?.Invoke();
            }
            Log.Warning(
                "For now, we are disabling the initialization process of Rocket Games Master if this is test automation build");
        }

        /// <summary>
        /// Calls given action if/when initialization is completed
        /// </summary>
        /// <param name="action"></param>
        public static void DoAfterInitialized(Action action)
        {
            if (IsInitialized)
            {
                action();
            }
            else
            {
                Action eventHandler = null;
                eventHandler = delegate
                {
                    if (!IsInitialized) return;
                    action();
                    Initialized -= eventHandler;
                };

                Initialized += eventHandler;
            }
        }

        protected virtual void OnApplicationPause(bool pauseStatus)
        {
            Log.Debug("OnApplicationPause: " + pauseStatus);

            if (!_firstFrame && pauseStatus)
            {
                RocGm.Config.LastActiveAt = DateTime.UtcNow;
            }

            //Don't fire when application is opening because event behaves differently on different platforms.
            if (!_firstFrame)
            {
                if (PauseStateChanged != null) PauseStateChanged(pauseStatus);
            }
            else
            {
                Log.Debug("On first frame. Will not fire Pause event.");
            }
        }

        protected virtual void OnApplicationFocus(bool hasFocus)
        {
            Log.Debug("OnApplicationFocus: " + hasFocus);

            //Don't fire when application is opening because event behaves differently on different platforms. 
            if (!_firstFrame)
            {
                if (FocusStateChanged != null) FocusStateChanged(hasFocus);
            }
            else
            {
                Log.Debug("On first frame. Will not fire Focus event.");
            }
        }

        protected virtual void OnApplicationQuit()
        {
            if (RocGm.Config != null)
                RocGm.Config.LastActiveAt = DateTime.UtcNow;

            if (ApplicationQuit != null) ApplicationQuit();
        }

        /// <summary>
        /// Is forced client update required?
        /// </summary>
        /// <returns>Returns boolean. It returns null if GM is not initialized.</returns>
        public bool? IsUpdateRequired()
        {
            if (!RocGm.IsInitialized)
            {
                Log.Warning("Rocket GM is not initialized yet!");
                return null;
            }

            Log.Debug("ClientVersion: " + RocGm.Config.ClientVersion.ToString("G", CultureInfo.InvariantCulture));
            Log.Debug("MinClientVersion: " + RocGm.GameSettings.MinClientVersion.ToString("G", CultureInfo.InvariantCulture));

#if ROC_DEBUG_MODE
            if (IsCustomBuildNumberUsageForced)
                return RocGm.GameSettings.MinClientVersion > CustomBuildNumber;
#endif

            return RocGm.GameSettings.MinClientVersion > RocGm.Config.ClientVersion;
        }

        protected virtual void OnQualitySettingsApplied(int qualitySetting)
        {
            float multiplier = (qualitySetting + 8) / 18f;

            Log.Debug("QualitySetting: " + qualitySetting + "Multiplier: " + multiplier, "ApplyQualitySettings");

            int width = Mathf.RoundToInt(RocGm.LiveTune.InitialWidth * multiplier);
            int height = Mathf.RoundToInt(RocGm.LiveTune.InitialHeight * multiplier);

            Screen.SetResolution(width, height, true);
            Log.Debug("Resolution is set. W: " + width + ", H: " + height);

            if (width != RocGm.LiveTune.InitialWidth || height != RocGm.LiveTune.InitialHeight)
            {
                Log.Debug("Resolution is changed!");

                if (ResolutionChanged != null) ResolutionChanged.Invoke(width, height);
            }
        }

        private bool RemoteCertificateValidationCallback(System.Object sender, X509Certificate certificate, X509Chain chain, SslPolicyErrors sslPolicyErrors)
        {
            bool isOk = true;
            // If there are errors in the certificate chain, look at each error to determine the cause.
            if (sslPolicyErrors != SslPolicyErrors.None)
            {
                for (int i = 0; i < chain.ChainStatus.Length; i++)
                {
                    if (chain.ChainStatus[i].Status != X509ChainStatusFlags.RevocationStatusUnknown)
                    {
                        chain.ChainPolicy.RevocationFlag = X509RevocationFlag.EntireChain;
                        chain.ChainPolicy.RevocationMode = X509RevocationMode.Online;
                        chain.ChainPolicy.UrlRetrievalTimeout = new TimeSpan(0, 1, 0);
                        chain.ChainPolicy.VerificationFlags = X509VerificationFlags.AllFlags;
                        bool chainIsValid = chain.Build((X509Certificate2)certificate);
                        if (!chainIsValid)
                        {
                            isOk = false;
                        }
                    }
                }
            }
            return isOk;
        }
    }
}
