﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using RocketCoroutine;
using RocketLog;
using UnityEngine;

namespace RocketGM
{
    public class RocGmFpsSampler
    {
        private readonly List<float> _snapshots = new List<float>();

        protected readonly RocLog Log = new RocLog(nameof(RocGmFpsSampler), DebugLevels.Info);

        /// <summary>
        /// Takes FPS Sample
        /// </summary>
        /// <param name="sampleCount">target number of samples</param>
        /// <param name="sampleRate">number of samples in a second</param>
        /// <param name="onComplete">returns average FPS</param>
        public void TakeSample(int sampleCount, int sampleRate, Action<float> onComplete)
        {
            _snapshots.Clear();
            
            FpsSampler(sampleCount, sampleRate, onComplete).StartCoroutine();
        }
        
        public IEnumerator FpsSampler(int targetSampleCount, int sampleRate, Action<float> onComplete)
        {

            while (targetSampleCount > _snapshots.Count)
            {
                yield return new WaitForSecondsRealtime(1.0f/sampleRate);

                var fps = 1.0f / Time.unscaledDeltaTime;
                _snapshots.Add(fps);
            }
            
            List<float> cleaned = SkipOutliers(_snapshots,1, result => result).ToList();

            float averageFps = cleaned.Average();
            onComplete(averageFps);
        }

        /// <summary>
        /// Calculates a standard deviation of elements, using a specified selector.
        /// </summary>
        public static double StandardDeviation<T>(IEnumerable<T> enumerable, Func<T, double> selector)
        {
            double sum = 0;
            var items = enumerable as T[] ?? enumerable.ToArray();
            double average = items.Average(selector);
            int n = 0;
            foreach (T item in items)
            {
                double diff = selector(item) - average;
                sum += diff * diff;
                n++;
            }
            return n == 0 ? 0 : Math.Sqrt(sum / n);
        }

        /// <summary>
        /// Filters elements to remove outliers. The enumeration will be
        /// selected three times, first to calculate an average, second
        /// for a standard deviation, and third to yield remaining elements. The outliers are these
        /// elements which are further from an average than k*(standard deviation). Set k=3 for
        /// standard three-sigma rule.
        /// </summary>
        public static IEnumerable<T> SkipOutliers<T>(IEnumerable<T> enumerable, double k, Func<T, double> selector)
        {
            // Duplicating a SD code to avoid calculating an average twice.
            double sum = 0;
            var items = enumerable as T[] ?? enumerable.ToArray();
            double average = items.Average(selector);
            int n = 0;
            foreach (T item in items)
            {
                double diff = selector(item) - average;
                sum += diff * diff;
                n++;
            }
            double sd = n == 0 ? 0 : Math.Sqrt(sum / n);
            double delta = k * sd;
            foreach (T item in items)
            {
                if (Math.Abs(selector(item) - average) <= delta)
                    yield return item;
            }
        }
    }
}
