﻿using System;
using System.Collections.Generic;
using System.Globalization;
using RocketGames;
using RocketGM.Models;
using RocketGM.RemoteSettings;
using RocketLog;
using UnityEngine;
using UnityEngine.Scripting;

#if UNITY_IOS
using RocketUtils.IDFAAuthorization;
using UnityEngine;
#endif

[assembly: AlwaysLinkAssembly]
namespace RocketGM
{
    public static class RocGmIDFAController
    {
        public static event Action<IDFAAuthorizationStatus> AuthorizationStatusChanged;
        public static event Action<bool, bool> IDFACustomUiAuthorizationInitiated;
        public static event Action<bool, bool, int> IDFANativeUiAuthorizationInitiated;

        public static bool IDFAAuthorizationRequested
        {
            get { return PlayerPrefs.GetInt(_idfaAuthorizationRequestedKey, 0) == 1; }
            private set { PlayerPrefs.SetInt(_idfaAuthorizationRequestedKey, value ? 1 : 0); }
        }

        public static bool IDFAAuthorizationInGameInitialShown
        {
            get { return PlayerPrefs.GetInt(_idfaAuthorizationInGameInitialShownKey, 0) == 1; }
            private set { PlayerPrefs.SetInt(_idfaAuthorizationInGameInitialShownKey, value ? 1 : 0); }
        }

        private static int IDFAAuthorizationTotalTryCount
        {
            get { return PlayerPrefs.GetInt(_idfaAuthorizationTotalTryCountKey, 0);  }
            set { PlayerPrefs.SetInt(_idfaAuthorizationTotalTryCountKey, value); }
        }

        private static string _idfaAuthorizationRequestedKey = "idfaAuthorizationRequested";
        private static string _idfaAuthorizationTotalTryCountKey = "idfaAuthorizationTotalTryCount";
        private static string _idfaAuthorizationInGameInitialShownKey = "idfaAuthorizationInGameInitialShown";

        private static RemoteInt _idfaAuthorizationInterval;
        private static RemoteInt _idfaAuthorizationInitialThreshold;
        private static RemoteBool _idfaAuthorizationEnabled;
        private static RemoteBool _idfaAuthorizationWithCustomUi;

        private static Action _callback;

#if UNITY_IOS

        private const string idfaAuthorizationStatusKey = "idfaAutharizationStatus";

        public static IDFAAuthorizationStatus IDFAAuthorizationStatus {
            private set
            {
                PlayerPrefs.SetInt(idfaAuthorizationStatusKey, (int) value);
            }
            get { return (IDFAAuthorizationStatus)PlayerPrefs.GetInt(idfaAuthorizationStatusKey, (int)IDFAAuthorizationStatus.NotDetermined); }
        }
#else
        public static IDFAAuthorizationStatus IDFAAuthorizationStatus = IDFAAuthorizationStatus.NotNecessary;
#endif

        private static RocLog _log = new RocLog("RocGmIDFAController", DebugLevels.Debug);
        

        [RuntimeInitializeOnLoadMethod]
        static void Initialize()
        {
#if ROC_GM_REPORTING
            _idfaAuthorizationInterval = new RemoteInt("idfaAuthorizationInterval", 3);
            _idfaAuthorizationInitialThreshold = new RemoteInt("idfaAuthorizationInitialThreshold", 1);
            _idfaAuthorizationEnabled = new RemoteBool("idfaAuthorizationEnabled", true);
            _idfaAuthorizationWithCustomUi = new RemoteBool("idfaAuthorizationWithCustomUi", false);
#endif
        }

        public static void ExecuteCallback()
        {
            if(_callback != null)
                _callback.Invoke();

            _callback = null;
        }

        public static void TryToInitiateAuthorization(bool overrideInterval = false, bool atStart = false, Action callback = null)
        {
            _callback = callback;

            if (!IsAuthorizationNeeded())
            {
                ExecuteCallback();
                return;
            }

            if (!_idfaAuthorizationEnabled || IDFAAuthorizationRequested)
            {
                ExecuteCallback();
                return;
            }

            if (overrideInterval)
            {
                TriggerRequestAuthorization(atStart, false);
                return;
            }

            IDFAAuthorizationTotalTryCount++;

            if (IDFAAuthorizationInGameInitialShown)
            {
                if (IDFAAuthorizationTotalTryCount % _idfaAuthorizationInterval == 0)
                {
                    TriggerRequestAuthorization(atStart, false);
                }
                else
                {
                    ExecuteCallback();
                }
            }
            else
            {
                if (IDFAAuthorizationTotalTryCount % _idfaAuthorizationInitialThreshold == 0)
                {
                    IDFAAuthorizationTotalTryCount = 0;
                    IDFAAuthorizationInGameInitialShown = true;
                    TriggerRequestAuthorization(atStart, true);
                }
                else
                {
                    ExecuteCallback();
                }
            }
        }

        private static void TriggerRequestAuthorization(bool atStart, bool isInitial)
        {
            if (_idfaAuthorizationWithCustomUi)
            {
                if (IDFACustomUiAuthorizationInitiated != null)
                    IDFACustomUiAuthorizationInitiated.Invoke(atStart, isInitial);
            }
            else
            {
                RequestAuthorization(atStart, isInitial);
            }
        }

        public static void RequestAuthorization(bool atStart = false, bool isInitial = false)
        {
            if(!_idfaAuthorizationEnabled || IDFAAuthorizationRequested) return;

            IDFAAuthorizationRequested = true;

            if (IsAuthorizationNeeded())
            {
#if UNITY_IOS
                int count = 0;

                if (atStart)
                {
                    count = 1;
                }
                else if (isInitial)
                {
                    count = RocGmController.Instance.IdfaAuthenticationOnInitialize ? 2 : 1;
                }
                else
                {
                    count = (IDFAAuthorizationTotalTryCount / _idfaAuthorizationInterval) + (RocGmController.Instance.IdfaAuthenticationOnInitialize ? 2 : 1);
                }

                if (IDFANativeUiAuthorizationInitiated != null)
                    IDFANativeUiAuthorizationInitiated.Invoke(atStart, isInitial, count);

                AdvertisingIdentifierController.ATTrackingManagerAuthorizationStatusEvent += OnATTrackingManagerAuthorizationStatusHandler;
                AdvertisingIdentifierController.RequestAuthorization();
#endif
            }
        }

        public static bool IsAuthorizationNeeded()
        {
#if UNITY_IOS
            if (AdvertisingIdentifierController.IsNeedToRequestIDFA())
            {
                return true;
            }
            else
            {
                _log.Debug("No need to request for IDFA");
#if UNITY_EDITOR
                return true;
#else
                return false;
#endif
            }
#else
            _log.Debug("Request Authorization is ONLY avaliable in iOS");
            return false;
#endif
        }



#if UNITY_IOS
        private static void OnATTrackingManagerAuthorizationStatusHandler(ATTrackingManagerAuthorizationStatus status)
        {
            AdvertisingIdentifierController.ATTrackingManagerAuthorizationStatusEvent -= OnATTrackingManagerAuthorizationStatusHandler;

            ExecuteCallback();

            IDFAAuthorizationStatus = (IDFAAuthorizationStatus)(int)status;

            RocGm.Config.AdvertisingId = IDFAAuthorizationStatus == IDFAAuthorizationStatus.Authorized ? AdvertisingIdentifierController.GetIDFA() : null;

            _log.Debug("**** IDFA: " + RocGm.Config.AdvertisingId);
            _log.Debug("**** IsAuthorizationNeeded: " + AdvertisingIdentifierController.IsNeedToRequestIDFA());
            _log.Debug("**** AuthorizationStatus: " + IDFAAuthorizationStatus);

            if (AuthorizationStatusChanged != null)
                AuthorizationStatusChanged.Invoke(IDFAAuthorizationStatus);
        }
#endif        
    }
}