﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Globalization;
using RocketCoroutine;
using RocketLog;
using RocketUtils;
using UnityEngine;
using UnityEngine.Networking;

namespace RocketGM
{
    public class RocGmLiveTune
    {
        public enum LiveTuneSource
        {
            None,
            Remote,
            Runtime
        }

        public event Action<int> QualitySettingsApplied;
        
        public bool IsLiveTuneEnabled
        {
            get
            {
                if (_isLiveTuneEnabled == -1)
                {
                    if (PlayerPrefs.HasKey("LiveTuneEnabled"))
                    {
                        _isLiveTuneEnabled = PlayerPrefs.GetInt("LiveTuneEnabled");
                    }
                    else
                    {
                        _isLiveTuneEnabled = 1;
                    }
                }
                return _isLiveTuneEnabled == 1 && _isLiveTuneEnabledFromGameSettings;
            }
            set
            {
                _isLiveTuneEnabled = value ? 1 : 0;
                PlayerPrefs.SetInt("LiveTuneEnabled", _isLiveTuneEnabled);
            }
        }

        public int InitialWidth
        {
            get { return _initialWidth; }
        }

        public int InitialHeight
        {
            get { return _initialHeight; }
        }

        public bool IsTuningAborted { get; private set; }
        public bool IsTuningCompleted { get; private set; }
        
        public event Action<int, int, LiveTuneSource> TuningCompleted;

        private const int SampleRate = 10;
        private const string QualitySettingKey = "qualitySetting";

        private int _isLiveTuneEnabled = -1;
	    private bool _isLiveTuneEnabledFromGameSettings;
        private LiveTuneSource _source = LiveTuneSource.None;
        private int _receivedQualitySetting;
        private int _targetFps = 60;
        private readonly Dictionary<int, float> _previousSamples = new Dictionary<int, float>();

        private readonly RocGmFpsSampler _fpsSampler = new RocGmFpsSampler();

        private int _initialWidth;
        private int _initialHeight;
        private int _initialFps = -1;

        private float _fpsOnPreviousSnapshot = -1;
        private float _fpsOnlastSnapshot = -1;
        private bool _waitingForApplyResult;
	    private bool _isConfigApplied;
        private int _previousQualitySetting = 10;
        private int _currentQualitySetting = 10;
	    private readonly int _minInitialFpsLimit = 20;

        protected readonly RocLog Log = new RocLog(nameof(RocGmLiveTune), DebugLevels.Debug);

        public RocGmLiveTune()
        {
	        RocGm.DoAfterInitialized((objects, settings) =>
	        {
		        _isLiveTuneEnabledFromGameSettings = settings.LiveTuneEnabled;

		        if (!IsLiveTuneEnabled) return;

		        _initialWidth = Screen.currentResolution.width;
		        _initialHeight = Screen.currentResolution.height;
		        _source = LiveTuneSource.None;
		        //DoAfterFixedUpdate to see logs 
		        CoroutineController.DoAfterFixedUpdate(LoadConfig);
	        });
        }

        #region Apply Config

        private void LoadConfig()
        {
	        _isConfigApplied = ApplyLocalConfig();
            if (_isConfigApplied) return;

            if (InternetConnectionController.HasConnection)
            {
                RocGm.DoAfterInitialized((parameters,gameSettings) =>
                {
	                _isConfigApplied = ApplyRemoteConfig(parameters);
                    if (!_isConfigApplied)
                    {
                        QuickTune();
                    };
                });
            }
            else
            {
                QuickTune();
            }
        }

        private bool ApplyRemoteConfig(Dictionary<string, object> parameters)
        {
            if (!parameters.ContainsKey(QualitySettingKey))
            {
                Log.Debug("Server did NOT return any QualitySetting");
                return false;
            }

            int qualitySetting;

            if (!int.TryParse(parameters[QualitySettingKey].ToString(), out qualitySetting))
            {
                Log.Error("Server returned MALFORMED QualitySetting");
                return false;
            }

            _receivedQualitySetting = qualitySetting;
            _source = LiveTuneSource.Remote;

            Log.Debug("Received QualitySetting from Server. QualitySetting: " + qualitySetting);
            ApplyConfig(qualitySetting);
            return true;
        }

        private bool ApplyLocalConfig()
        {
            if (PlayerPrefs.GetInt(QualitySettingKey, -1) < 0) return false;

            int qualitySetting = PlayerPrefs.GetInt(QualitySettingKey);

            _receivedQualitySetting = qualitySetting;

	        IsTuningCompleted = true;

            Log.Debug("Local Config will be applied. QualitySetting: " + qualitySetting);

            return ApplyConfig(qualitySetting);
        }

        private bool ApplyConfig(int qualitySetting)
        {
			Log.Info("Config will be applied. QualitySetting: " + qualitySetting);

            //Apply resolutionMultiplier IMMEDIATELY!
            SetQualitySetting(qualitySetting);

            return true;
        }
        #endregion

        private bool CanBeTuned()
        {
            if (!IsLiveTuneEnabled)
            {
                Log.Debug("LiveTune is DISABLED");
                return false;
            }

            if (IsTuningCompleted)
            {
                Log.Debug("Tuning is already completed");
                return false;
            }

            if (_waitingForApplyResult)
            {
                Log.Warning("There is unapplied results. Snapshot will NOT be taken");
                return false;
            }

            return true;
        }

        /// <summary>
        /// Does a quick tune which does not affect normal tuning process
        /// May cause flicker!
        /// </summary>
        public void QuickTune()
        {
            if (!CanBeTuned()) return;

            Log.Debug("QuickTune is started. Taking sample");

            _fpsSampler.TakeSample(30, 15, averageFps =>
            {
                RecordSample(_currentQualitySetting, averageFps);

                if (averageFps / _targetFps > 0.75f)
                {
                    Log.Debug("Average FPS: " + averageFps + ", No need to QuickTune!");
                    IsTuningAborted = true;
                    return;
                }

	            _initialFps = (int)averageFps;

				int targetQualitySetting = CalculateQualitySetting(averageFps, _targetFps);
                Log.Debug("QuickTune is finished. Average FPS: " + averageFps + ", QualitySetting: " + targetQualitySetting);

                SetQualitySetting(targetQualitySetting);
            });
        }

        public void TakeSnapshot(int duration, bool applyOnCompleted = false)
        {
            if (!CanBeTuned()) return;

            Log.Debug("Taking Snapshot");

            _fpsSampler.TakeSample(duration * SampleRate, SampleRate, averageFps =>
            {
                Log.Debug("## Average FPS: " + averageFps);
                RecordSample(_currentQualitySetting, averageFps);

                if (averageFps / _targetFps > 0.75f)
                {
                    Log.Debug("Average FPS: " + averageFps + ", No need to Apply Result!");
                    IsTuningAborted = true;
                    return;
                }

                if (_fpsOnlastSnapshot < float.Epsilon && _initialFps == -1)
                {
                    _initialFps = (int) averageFps;

	                Log.Debug("Initial FPS: " + _initialFps);
                }

                _fpsOnPreviousSnapshot = _fpsOnlastSnapshot;
                _fpsOnlastSnapshot = averageFps;
                _waitingForApplyResult = true;
                
                if (applyOnCompleted)
                    ApplyResult();
            });
        }

        public void ApplyResult()
        {
            if (IsTuningCompleted)
            {
                Log.Debug("Tuning is already completed");
                return;
            }

            if (!_waitingForApplyResult)
            {
                Log.Debug("No Snapshot to Apply");
                return;
            }

	        if (_isConfigApplied && _fpsOnPreviousSnapshot < 0)
	        {
				Log.Debug("Config applied but previous snapshot is NOT available. Wait for next capture.");
		        _waitingForApplyResult = false;
				return;
	        }

	        _waitingForApplyResult = false;

			if (_fpsOnlastSnapshot <= _minInitialFpsLimit || (_currentQualitySetting <= 5 && _fpsOnlastSnapshot <= _minInitialFpsLimit*2))
			{
				_targetFps /= 2;
			}

			float increaseInLastTry = (_fpsOnlastSnapshot - _fpsOnPreviousSnapshot) / _targetFps;
            float fpsToTarget = _fpsOnlastSnapshot / _targetFps;

	        Log.Debug(string.Format("FpsToTarget: {0}," +
	                                "\nIncreaseInLastTry: {1}," +
	                                "\nFpsOnPreviousSnapshot: {2}," +
	                                "\nCurrentQualitySetting: {3}",
		        fpsToTarget, increaseInLastTry, _fpsOnPreviousSnapshot, _currentQualitySetting));

			if (fpsToTarget > 0.9f)
            {
	            if (fpsToTarget > 0.95 && increaseInLastTry > 0.3 && _fpsOnPreviousSnapshot >= 0)
                {
                    Log.Debug("Target FPS is reached. But There was too big difference try higher value");
                    int targetQualitySetting = Mathf.RoundToInt((_currentQualitySetting + _previousQualitySetting) / 2f);
                    if (targetQualitySetting == _currentQualitySetting)
                    {
                        targetQualitySetting = _currentQualitySetting + 1;
                    }
                    SetQualitySetting(targetQualitySetting);
                }
                else
                {
                    Log.Debug("Target FPS is reached");
                    SetFinalQualitySetting(_currentQualitySetting);
                }
                return;
            }

            if (Math.Abs(increaseInLastTry) < 0.05)
            {
                Log.Debug("Nothing changed on last try, stay! targetFrameRate:" + _fpsOnlastSnapshot);                
                SetFinalQualitySetting(_currentQualitySetting);
            }
            else if (increaseInLastTry < -0.2)
            {
                Log.Debug("Gone too bad, try lower value");
                int targetQualitySetting = Mathf.RoundToInt((_currentQualitySetting + _previousQualitySetting) / 2f);
                if (targetQualitySetting == _currentQualitySetting)
                {
                    targetQualitySetting = _currentQualitySetting - 1;
                }

                SetQualitySetting(targetQualitySetting);
            }
            else if (increaseInLastTry < -0.05)
            {
                Log.Debug("Gone a little bad, go back and stay");
                SetQualitySetting(_previousQualitySetting);
                SetFinalQualitySetting(_currentQualitySetting);
            }
            else
            {
                Log.Debug("Got better");
	            int targetQualitySetting = CalculateQualitySetting(_fpsOnlastSnapshot, _targetFps);

				if (targetQualitySetting >= _currentQualitySetting)
				{
					targetQualitySetting = _currentQualitySetting - 1;
				}

				SetQualitySetting(targetQualitySetting);
            }
        }

        public void SetTargetFPS(int fps)
        {
            _targetFps = fps;
            Log.Warning($"WARNING: Target FPS is set to {fps}");
        }

        private int CalculateQualitySetting(float averageFps, float targetFps)
        {
            return Mathf.Clamp(Mathf.RoundToInt(((averageFps * 9f) / targetFps) + 1), 1, 10);
        }

        protected virtual void SetQualitySetting(int qualitySetting)
        {
	        qualitySetting = Mathf.Clamp(qualitySetting, 1, 10);

            if (_currentQualitySetting == qualitySetting)
            {
                ApplyQualitySetting(_currentQualitySetting);
                return;
            }

            Action doAfterApply = null;
            if (_previousSamples.ContainsKey(qualitySetting))
            {
                Log.Debug("We are in a loop, choose best sample and continue");

                KeyValuePair<int, float> bestSample = new KeyValuePair<int, float>(0, 0);
                foreach (KeyValuePair<int, float> sample in _previousSamples)
                {
                    if (sample.Value > bestSample.Value)
                    {
                        bestSample = sample;
                    }
                }

                Log.Debug("## BEST SAMPLE: QualitySetting: " + bestSample.Key + ", FPS: " + bestSample.Value);

                qualitySetting = bestSample.Key;
                _fpsOnlastSnapshot = bestSample.Value;

                doAfterApply = () => SetFinalQualitySetting(qualitySetting);
            }

            Log.Debug("Set QualitySetting: " + qualitySetting);

            _previousQualitySetting = _currentQualitySetting;
            _currentQualitySetting = qualitySetting;
        
            ApplyQualitySetting(_currentQualitySetting);

            if (doAfterApply != null) doAfterApply();
        }

        private void ApplyQualitySetting(int qualitySetting)
        {
            Application.targetFrameRate = _targetFps;
            QualitySettings.vSyncCount = 0;

            if (QualitySettingsApplied != null) QualitySettingsApplied(qualitySetting);
        }

        private void SetFinalQualitySetting(int qualitySetting)
        {
            IsTuningCompleted = true;
            Log.Info("Best value for Quality Setting is: " + qualitySetting + ", best FPS available: " + _fpsOnlastSnapshot);
            PlayerPrefs.SetInt(QualitySettingKey, qualitySetting);

	        Application.targetFrameRate = (int)_fpsOnlastSnapshot;

            if (_receivedQualitySetting != qualitySetting)
            {
                _source = LiveTuneSource.Runtime;
            }

			ReportPerformance(qualitySetting, _initialFps, (int)_fpsOnlastSnapshot);
            if (TuningCompleted != null) TuningCompleted(qualitySetting, (int) _fpsOnlastSnapshot, _source);
        }

        public void ReportPerformance(int qualitySetting, int initialFps, int finalFps)
        {
            ReportPerformanceRoutine(qualitySetting, initialFps, finalFps).StartCoroutine();
        }

        private IEnumerator ReportPerformanceRoutine(int qualitySetting, int initialFps, int finalFps)
        {
	        if (!RocGm.IsInitialized)
	        {
		        Log.Warning("RocGm is NOT Initialized yet!");
				yield break;
			}	        

            Log.Debug(string.Format("PerformanceReport:" +
                                     "\nplayerId:{0}" +
                                     "\ngameId: {1}" +
                                     "\nqualitySetting: {2}" +
                                     "\ninitialFps: {3}" +
                                     "\nfinalFps: {4}",
                RocGm.PlayerId, RocGm.Config.GameId, qualitySetting, initialFps, finalFps));


			WWWForm request = RocGm.CreateRequest();
			
			request.AddField("deviceModel", RocGm.Config.DeviceModel);
            request.AddField("deviceOs", RocGm.Config.DeviceOs);
            request.AddField("qualitySetting", qualitySetting.ToString(CultureInfo.InvariantCulture));
            request.AddField("initialFps", initialFps.ToString());
            request.AddField("finalFps", finalFps.ToString());

            UnityWebRequest www = UnityWebRequest.Post(string.Format("{0}/performance", RocGm.RocGmTrackerUrl), request);

	        www.chunkedTransfer = false;
	        www.useHttpContinue = false;

			yield return www.SendWebRequest();

            if (www.responseCode != 200)
            {
                Log.Error(string.Format("Performance report REJECTED! responseCode: {0} errorCode: {1}", www.responseCode, www.downloadHandler.text));
                yield break;
            }

            Dictionary<String, System.Object> result;
            if (!RocGmHelper.TryGetServerResult(www.downloadHandler.text, out result))
            {
                yield break;
            };

            Log.Info("Performance report is sent!");
        }

        private void RecordSample(int qualitySetting, float averageFps)
        {
            if (_previousSamples.ContainsKey(qualitySetting))
            {
                Log.Warning("Sample is ALREADY recorded");
                return;
            }

            _previousSamples.Add(qualitySetting, averageFps);
        }

    }
}
