﻿using System;
using System.Collections;
using System.Collections.Generic;
using RocketCoroutine;
using RocketLog;
using RocketUtils.SerializableDictionary;
using UnityEngine;
using UnityEngine.Networking;

namespace RocketGM
{
    public class RocGmPlayerProgress
    {
        protected readonly RocLog Log = new RocLog(nameof(RocGm), DebugLevels.Debug);

        private const string PlayerProgressKey = "gmPlayerProgress";
        private const string NonFinalizedPlayerProgressKey = "gmNonFinalizedPlayerProgress";
        private const string RetryCoroutineKey = "gmRetrySendProgress";
        private const string SendCoroutineKey = "gmSendProgress";
        private const string GameStartKey = "gmProgressGameStart";
        private const string ProgressLevel = "gmProgressLevel";

        private bool _gameStartTracked
        {
            get { return PlayerPrefs.GetInt(GameStartKey, 0) == 1; }
            set { PlayerPrefs.SetInt(GameStartKey, Convert.ToInt32(value)); }
        }

        private int _progressLevel
        {
            get { return PlayerPrefs.GetInt(ProgressLevel, 1); }
            set { PlayerPrefs.SetInt(ProgressLevel, value); }
        }

        private const float RetryInterval = 10f;

        private readonly SerializableDictionary<string, PlayerProgress> _nonFinalizedProgressDictionary;
        private readonly SerializableDictionary<string, PlayerProgress> _finalizedPlayerProgressDictionary;
        private readonly List<PlayerProgress> _waitingToBeSentProgressList;

        private bool _isTrackProgressActive;

        private bool _isPersistent;

        [Serializable]
        private class PlayerProgressCollection
        {
            public List<PlayerProgress> ProgressList = new List<PlayerProgress>();
        }

        private readonly PlayerProgressCollection _playerProgressCollection = new PlayerProgressCollection();

        public RocGmPlayerProgress(bool persistentProgressEnabled)
        {
            _isPersistent = persistentProgressEnabled;

            RocGm.DoAfterInitialized((objects, settings) => _isTrackProgressActive = settings.TrackProgress);            

            RocGmControllerBase.PauseStateChanged += OnPauseStateChanged;
            RocGmControllerBase.FocusStateChanged += OnFocusStateChanged;

            _finalizedPlayerProgressDictionary = new SerializableDictionary<string, PlayerProgress>();
            _nonFinalizedProgressDictionary = new SerializableDictionary<string, PlayerProgress>();
            _waitingToBeSentProgressList = new List<PlayerProgress>();

            if (!_gameStartTracked)
            {
                StartProgress("GameStart");
            }

            if (PlayerPrefs.HasKey(PlayerProgressKey))
            {
                var progressString = PlayerPrefs.GetString(PlayerProgressKey);
                _finalizedPlayerProgressDictionary = JsonUtility.FromJson<SerializableDictionary<string, PlayerProgress>>(progressString);

                SendProgress();
            }

            if (_isPersistent )
            {
                if (PlayerPrefs.HasKey(NonFinalizedPlayerProgressKey))
                {
                    var nonFinalizedProgressString = PlayerPrefs.GetString(NonFinalizedPlayerProgressKey);
                    _nonFinalizedProgressDictionary = JsonUtility.FromJson<SerializableDictionary<string, PlayerProgress>>(nonFinalizedProgressString);
                }
            }
        }

        #region Event Listeners
        private void OnFocusStateChanged(bool isFocused)
        {
            if (isFocused)
            {
                ResumeAllProgresses();
            }
            else
            {
                PauseAllProgresses();
            }
        }


        private void OnPauseStateChanged(bool isPaused)
        {
            if (isPaused)
            {
                PauseAllProgresses();                
            }
            else
            {
                ResumeAllProgresses();
            }
        }
        #endregion

        #region Progress Methods

        public void StartProgress(string stepName)
        {
            if (string.IsNullOrEmpty(stepName))
            {
                Log.Error("Progress can not be started, name is empty or null");
                return;
            }
            
            if (!_gameStartTracked && !stepName.Equals("GameStart"))
            {
                _gameStartTracked = true;

                StopProgress("GameStart", 1);
            }
            
            PlayerProgress playerProgress = new PlayerProgress(stepName);
            Log.Debug("Progress Started: " + stepName);
            AddProgressToNonFinalizedDictionary(playerProgress);
        }
        
        private void StartProgressWithDuration(string stepName, float duration)
        {            
            PlayerProgress playerProgress = new PlayerProgress(stepName, duration);
            Log.Debug("Progress Started: " + stepName);
            AddProgressToNonFinalizedDictionary(playerProgress);
        }

        /// <summary>
        /// Starts new progress. Level logic is handled automatically
        /// </summary>
        public void StartProgress()
        {
            StartProgress(_progressLevel);
        }

        /// <summary>
        /// Starts new progress for given level. Use this if you want to send level manually, otherwise don't send parameter to let GM handle level logic.
        /// </summary>
        /// <param name="level"></param>
        public void StartProgress(int level)
        {
            if (level < 1)
            {
                Log.Error("Level can NOT be lower than 1!");
                return;
            }

            string stepName = "L" + level.ToString("D3");
            StartProgress(stepName);
        }
        
        
        public void StartProgressWithDuration(float duration)
        {
            StartProgressWithDuration(_progressLevel, duration);
        }
        
        public void StartProgressWithDuration(int level, float duration)
        {
            if (level < 1)
            {
                Log.Error("Level can NOT be lower than 1!");
                return;
            }

            if (!_gameStartTracked)
            {
                _gameStartTracked = true;

                StopProgress("GameStart", 1);
            }

            string stepName = "L" + level.ToString("D3");
            StartProgressWithDuration(stepName, duration);
        }

        public void PauseProgress(string stepName)
        {
            PlayerProgress playerProgress = GetNonFinalizedProgress(stepName);
            if (playerProgress == null)
            {
                Log.Warning("Player progress with key " + stepName + " does NOT exist!", "PauseProgress");
                return;
            }

            Log.Debug("Progress paused:" + stepName);

            playerProgress.Pause();
        }
        
        /// <summary>
        /// Pause current progress. Level logic is handled automatically
        /// </summary>
        public void PauseProgress()
        {
            PauseProgress(_progressLevel);
        }

        /// <summary>
        /// Pause progress for given level. Use this if you want to send level manually, otherwise don't send parameter to let GM handle level logic.
        /// </summary>
        /// <param name="level"></param>
        public void PauseProgress(int level)
        {
            if (level < 1)
            {
                Log.Warning("Level can NOT be lower than 1!");
                return;
            }

            string stepName = "L" + level.ToString("D3");
            PauseProgress(stepName);
        }

        public void ResumeProgress(string stepName)
        {
            PlayerProgress playerProgress = GetNonFinalizedProgress(stepName);
            if (playerProgress == null)
            {
                Log.Warning("Player progress with key " + stepName + " does NOT exist!", "ResumeProgress");
                return;
            }

            Log.Debug("Progress resumed:" + stepName);
            playerProgress.Resume();
        }

        /// <summary>
        /// Resume previously paused progress. Level logic is handled automatically
        /// </summary>
        public void ResumeProgress()
        {
            ResumeProgress(_progressLevel);
        }

        /// <summary>
        /// Resume progress for given level. Use this if you want to send level manually, otherwise don't send parameter to let GM handle level logic.
        /// </summary>
        /// <param name="level"></param>
        public void ResumeProgress(int level)
        {
            if (level < 1)
            {
                Log.Error("Level can NOT be lower than 1!");
                return;
            }

            string stepName = "L" + level.ToString("D3");
            ResumeProgress(stepName);
        }

        public void StopProgress(string stepName, int status)
        {
            if (status < 0 || status > 2)
            {
                Log.Error("Stop progress could not be initiated. Please send one of the following options as status parameter:" +
                          "\n- 0: Player failed the level." +
                          "\n- 1: Player succeeded the level." +
                          "\n- 2: Player failed the level, but still increase the player's level in GM.");
                return;
            }
            
            PlayerProgress playerProgress = GetNonFinalizedProgress(stepName);
            if (playerProgress == null)
            {
                Log.Warning("Player progress with key " + stepName + " does NOT exist!", "StopProgress");
                return;
            }            
            
            Log.Debug(string.Format("Progress Stopped.Name:{0} Status:{1}", stepName, status));

            var finalizeStatus = status == 1 ? 1 : 0;
            
            playerProgress.Finalize(finalizeStatus);

            _nonFinalizedProgressDictionary.Remove(playerProgress.stepName);            

            SendProgress(playerProgress);

            if ((status == 1 || status == 2) && stepName.Contains("L"))
                _progressLevel++;
        }

        /// <summary>
        /// Status( 0->Fail,1->Success,2->FailWithProgress ). Stops current progress and sends it to server. Level logic is handled automatically
        /// </summary>
        /// <param name="status"></param>
        public void StopProgress(int status)
        {
            StopProgress(_progressLevel, status);
        }

        /// <summary>
        /// Status( 0->Fail,1->Success,2->FailWithProgress ). Stops progress for given level and sends it to server. Level logic is handled automatically
        /// </summary>
        /// <param name="level"></param>
        /// <param name="status"></param>
        public void StopProgress(int level, int status)
        {
            if (level < 1)
            {
                Log.Error("Level can NOT be lower than 1!");
                return;
            }

            string stepName = "L" + level.ToString("D3");

            StopProgress(stepName, status);
        }
        #endregion

        #region Send Progress
        private void SendProgress(PlayerProgress playerProgress)
        {
            if (CoroutineController.IsCoroutineRunning(SendCoroutineKey))
            {
                _waitingToBeSentProgressList.Add(playerProgress);
                return;
            }

            AddProgressToFinalizedDictionary(playerProgress);

            SendProgress();
        }

        private void SendProgress()
        {
            _playerProgressCollection.ProgressList = GetPlayerProgressList();

            if (_playerProgressCollection.ProgressList.Count == 0)
            {
                Log.Warning("PlayerProgressCollection is EMPTY", "SendProgress");
                return;
            }

            SendProgress(JsonUtility.ToJson(_playerProgressCollection));
        }

        private void SendProgress(string playerProgress)
        {
            CoroutineController.StartCoroutine(SendProgressRoutine(playerProgress), SendCoroutineKey);
        }

        private IEnumerator SendProgressRoutine(string playerProgress)
        {
            if (!RocGm.IsInitialized)
            {
                Log.Warning("RocGm is NOT Initialized yet!");
                RocGm.DoAfterInitialized(SendProgress);
                SaveProgress();
                yield break;
            }

            //If tracking is disabled, clear everything.
            if (!_isTrackProgressActive)
            {
                Log.Warning("Player Progress is disabled in GM. Please enable it.");
                _waitingToBeSentProgressList.Clear();
                _nonFinalizedProgressDictionary.Clear();
                _finalizedPlayerProgressDictionary.Clear();
                DeleteProgress();
                yield break;
            }

            string simplifiedPlayerProgress = playerProgress.Replace("\"ProgressList\":", "");
            simplifiedPlayerProgress = simplifiedPlayerProgress.Substring(1, simplifiedPlayerProgress.Length - 2);

            WWWForm request = RocGm.CreateRequest();

            request.AddField("playerId", RocGm.PlayerId.ToString());
            request.AddField("progress", simplifiedPlayerProgress);

            using var www = UnityWebRequest.Post(string.Format("{0}/updatePlayerProgress", RocGm.RocGmTrackerUrl), request);
            www.chunkedTransfer = false;
            www.useHttpContinue = false;
            www.timeout = 30;

            Log.Debug("Sending UpdatePlayerProgress request");

            yield return www.SendWebRequest();
            Dictionary<string, object> parameters;
            RocGmHelper.TryGetServerResult(www.downloadHandler.text, out parameters);

            bool shouldResend = parameters == null || (parameters.ContainsKey("ServerReturnCode") && Convert.ToInt32(parameters["ServerReturnCode"]) == -1);

            if (shouldResend)
            {
                Log.Warning(string.Format("UpdatePlayerProgress REJECTED! responseCode: {0} errorCode: {1}", www.responseCode, www.downloadHandler.text));

                SaveProgress();
                CoroutineController.DoAfterGivenTime(RetryInterval, SendProgress, RetryCoroutineKey, true);
            }
            else
            {
                if (www.responseCode != 200)
                {
                    Log.Warning(string.Format("UpdatePlayerProgress REJECTED! responseCode: {0} errorCode: {1}", www.responseCode, www.downloadHandler.text));
                }

                _finalizedPlayerProgressDictionary.Clear();
                DeleteProgress();
            }

            //If there are progresses that are waiting to be sent, add them to finalized dictionary
            foreach (var waitingProgress in _waitingToBeSentProgressList)
            {
                Log.Debug("Adding Waiting Progress: " + waitingProgress.Key + " to Finalized Dictionary");
                AddProgressToFinalizedDictionary(waitingProgress);
            }

            _waitingToBeSentProgressList.Clear();

            if (!shouldResend && _finalizedPlayerProgressDictionary.Count > 0)
                CoroutineController.DoAfterFixedUpdate(SendProgress);
        }
        #endregion

        #region Helpers
        private void PauseAllProgresses()
        {
            if (_isPersistent)
            {
                SaveNonFinalizedProgress();
            }

            foreach (var playerProgress in _nonFinalizedProgressDictionary.Values)
            {
                playerProgress.Pause();
            }
        }

        private void ResumeAllProgresses()
        {
            foreach (var playerProgress in _nonFinalizedProgressDictionary.Values)
            {
                playerProgress.Resume();
            }
        }

        private List<PlayerProgress> GetPlayerProgressList()
        {
            List<PlayerProgress> playerProgresses = new List<PlayerProgress>();

            foreach (PlayerProgress playerProgress in _finalizedPlayerProgressDictionary.Values)
            {
                playerProgresses.Add(playerProgress);
            }

            return playerProgresses;
        }

        private void AddProgressToNonFinalizedDictionary(PlayerProgress playerProgress)
        {
            if (_nonFinalizedProgressDictionary.ContainsKey(playerProgress.stepName))
            {
                if (!_isPersistent)
                {
                    Log.Warning("Same progress with key: " + playerProgress.stepName + "is ALREADY being tracked!");
                }
                else
                {
                    PlayerProgress nonFinalizedProgress = _nonFinalizedProgressDictionary[playerProgress.stepName];
                    nonFinalizedProgress.Resume();
                }
                return;
            }

            _nonFinalizedProgressDictionary.Add(playerProgress.stepName, playerProgress);
        }

        private void AddProgressToFinalizedDictionary(PlayerProgress playerProgress)
        {
            if (!_finalizedPlayerProgressDictionary.ContainsKey(playerProgress.Key))
            {
                _finalizedPlayerProgressDictionary.Add(playerProgress.Key, playerProgress);
            }
            else
            {
                PlayerProgress savedPlayerProgress = _finalizedPlayerProgressDictionary[playerProgress.Key];
                savedPlayerProgress.stepCount += playerProgress.stepCount;
                savedPlayerProgress.stepDuration += playerProgress.stepDuration;
            }
        }

        private PlayerProgress GetNonFinalizedProgress(string stepName)
        {
            if (!_nonFinalizedProgressDictionary.ContainsKey(stepName))
            {
                Log.Warning("Player progress with key " + stepName + " does NOT exist!", "GetNonFinalizedProgress");
                return null;
            }

            return _nonFinalizedProgressDictionary[stepName];
        }

        private void SaveProgress()
        {
            PlayerPrefs.SetString(PlayerProgressKey, JsonUtility.ToJson(_finalizedPlayerProgressDictionary));
        }

        private void SaveNonFinalizedProgress()
        {
            foreach (PlayerProgress playerProgress in _nonFinalizedProgressDictionary.Values)
            {
                playerProgress.SaveDuration();
            }

            PlayerPrefs.SetString(NonFinalizedPlayerProgressKey, JsonUtility.ToJson(_nonFinalizedProgressDictionary));
        }

        private void DeleteProgress()
        {
            if (PlayerPrefs.HasKey(PlayerProgressKey))
                PlayerPrefs.DeleteKey(PlayerProgressKey);

            if (_isPersistent)
            {
                if (PlayerPrefs.HasKey(NonFinalizedPlayerProgressKey))
                    PlayerPrefs.DeleteKey(NonFinalizedPlayerProgressKey);
            }
        }
        #endregion
    }
}
