using System;
using System.Collections.Generic;
using Newtonsoft.Json;
using RocketLog;
using UnityEngine.Serialization;

namespace RocketGM
{
    [Serializable]
    public class RocGmPlayerProperty
    {
        public int playerPropertyId;
        public string playerPropertyKey;
        public string playerPropertyValue;
        public string playerPropertyDataType;
    }

    [Serializable]
    public class RocGmPlayerProperties
    {
        private Dictionary<string, RocGmPlayerProperty> CurrentPlayerProperties;
        private RocLog Log = new RocLog(nameof(RocGmPlayerProperties), DebugLevels.Debug);

        // For JSON serialization.
        internal RocGmPlayerProperties()
        {
        }

        internal void CopyFrom(List<RocGmPlayerProperty> availablePlayerProperties,
            List<RocGmPlayerProperty> currentPlayerProperties)
        {
            CurrentPlayerProperties = new Dictionary<string, RocGmPlayerProperty>();
            foreach (var currentProperty in currentPlayerProperties)
            {
                CurrentPlayerProperties.Add(currentProperty.playerPropertyKey, currentProperty);
            }

            foreach (var propertyTemplate in availablePlayerProperties)
            {
                if (CurrentPlayerProperties.ContainsKey(propertyTemplate.playerPropertyKey))
                    continue;
                
                CurrentPlayerProperties.Add(propertyTemplate.playerPropertyKey, new RocGmPlayerProperty()
                {
                    playerPropertyId = propertyTemplate.playerPropertyId,
                    playerPropertyValue = null,
                    playerPropertyKey = propertyTemplate.playerPropertyKey,
                    playerPropertyDataType = propertyTemplate.playerPropertyDataType
                });
            }
        }

        public bool TrySetPlayerProperty(string key, object value)
        {
            if (!CurrentPlayerProperties.ContainsKey(key))
            {
                Log.Error($"User property with key {key} does not exist in {nameof(CurrentPlayerProperties)}.");
                return false;
            }

            if (!CurrentPlayerProperties.TryGetValue(key, out var userProperty)) return false;

            switch (value)
            {
                case bool _:
                    if (userProperty.playerPropertyDataType != "boolean")
                    {
                        Log.Error($"Property is not a valid boolean.");
                        return false;
                    }

                    break;
                case int _:
                    if (userProperty.playerPropertyDataType != "int")
                    {
                        Log.Error($"Property is not a valid integer.");
                        return false;
                    }

                    break;
                case string _:
                    if (userProperty.playerPropertyDataType != "string" && userProperty.playerPropertyDataType != "json")
                    {
                        Log.Error($"Property is not a valid string.");
                        return false;
                    }

                    break;
                default:
                    Log.Error($"Property is not a valid type.");
                    return false;
            }


            userProperty.playerPropertyValue = value.ToString();
            RocGm.SendPlayerProperty(userProperty.playerPropertyId, userProperty.playerPropertyValue);

            return true;
        }

        #region Getters With Type Casts

        // public bool TryGetUserProperty(string key, out object value)
        // {
        //     value = null;
        //
        //     if (!CurrentUserProperties.TryGetValue(key, out var userProperty)) return false;
        //     
        //     switch (value)
        //     {
        //         case bool:
        //             if (userProperty.userPropertyDataType != "boolean")return false;
        //             value = userProperty.userPropertyValue == "true" || userProperty.userPropertyValue == "1";
        //             return true;
        //         case int:
        //             if (userProperty.userPropertyDataType != "int") return false;
        //             value = int.Parse(userProperty.userPropertyValue);
        //             return true;
        //         case string:
        //             if (userProperty.userPropertyDataType != "string") return false;
        //             value = userProperty.userPropertyValue;
        //             return true;
        //         default:
        //             return false;
        //     }
        //
        //     return false;
        // }

        public bool TryGetPlayerProperty(string key, out string value)
        {
            value = null;

            if (!CurrentPlayerProperties.TryGetValue(key, out var userProperty)) return false;

            value = userProperty.playerPropertyValue;

            return true;
        }

        public bool TryGetPlayerProperty(string key, out bool value)
        {
            value = false;

            if (!CurrentPlayerProperties.TryGetValue(key, out var userProperty)) return false;

            value = bool.Parse(userProperty.playerPropertyValue);

            return true;
        }

        public bool TryGetPlayerProperty(string key, out int value)
        {
            value = 0;

            if (!CurrentPlayerProperties.TryGetValue(key, out var userProperty)) return false;

            value = int.Parse(userProperty.playerPropertyValue);

            return true;
        }

        public bool TryGetPlayerProperty<T>(string key, out T value)
        {
            value = default;

            if (!CurrentPlayerProperties.TryGetValue(key, out var userProperty)) return false;

            value = JsonConvert.DeserializeObject<T>(userProperty.playerPropertyValue);

            return true;
        }

        #endregion
    }
}