﻿using JetBrains.Annotations;
using RocketCoroutine;
using RocketGM.Models;
using RocketLog;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Globalization;
using RocketGM.Models.UpdateSession;
using RocketUtils.Models;
using UnityEngine;
using UnityEngine.Networking;

namespace RocketGM
{
    public class RocGmReporting
    {
	    public bool IsPaymentInProgress => CoroutineController.IsCoroutineRunning(_paymentRoutine);

	    private const int _retryInterval = 10;

        private static Dictionary<string, object> _appsFlyerAttributionParameters;
        
        private IEnumerator _paymentRoutine;

        private readonly RocGmSessionReporting _rocGmSessionReporting;

        private readonly RocLog Log = new RocLog(nameof(RocGm), DebugLevels.Debug);

        internal RocGmReporting()
        {
            RocGmConfig.AdvertisingIdReceived += OnAdvertisingIdReceived;

            if (RocGmIDFAController.IDFAAuthorizationStatus == IDFAAuthorizationStatus.NotDetermined)
            {
                RocGmIDFAController.AuthorizationStatusChanged += OnAuthorizationStatusChanged;
            }

            _rocGmSessionReporting = new RocGmSessionReporting();
        }
        
#if ROC_AD
        public void AdStarted()
        {
	        _rocGmSessionReporting.OnAdStarted();
        }
        public void AdFinished()
        {
	        _rocGmSessionReporting.OnAdFinished();
        }
#endif

        public void SaveAdWatched(ImpressionData impressionData)
        {
	        _rocGmSessionReporting.SaveAdWatched(impressionData);
        }

        public void SaveAdWatched(AdFormat adFormat)
        {
	        _rocGmSessionReporting.SaveAdWatched(adFormat);
        }

        public void UpdateEstimatedAdRevenue(double revenue, AdFormat adFormat)
        {
	        _rocGmSessionReporting.UpdateEstimatedAdRevenue(revenue, adFormat);
        }

        #region Payment

        /// <summary>
        /// Check payment receipt and log revenue
        /// </summary>
        /// <param name="isSubscription">subscription: true, iap: false</param>
        /// <param name="receipt">receipt taken from app store / google play</param>
        /// <param name="userInitiated">is user initiated payment process?</param>
        /// <param name="isConsumable"></param>
        /// <param name="price">price</param>
        /// <param name="currency">3 digit currency i.e. usd, try, eur etc.</param>
        /// <param name="productName"></param>
        /// <param name="originalTransactionId"></param>
        /// <param name="callback"></param>
        public void Payment(bool isSubscription, string receipt, bool userInitiated, bool isConsumable,
			decimal? price = null,
			string currency = null, string productName = null, string originalTransactionId = null,
			Action<bool, Dictionary<String, System.Object>> callback = null)
		{
            RocGm.DoAfterInitialized(() =>
            {
                if (string.IsNullOrEmpty(receipt))
                {
                    Log.Error("Receipt is EMPTY!");
                    if (callback != null) callback(false, null);
                    return;
                }

                Log.Debug("Payment is started");

                CoroutineController.StartCoroutine(_paymentRoutine = PaymentRoutine(isSubscription, receipt,
	                userInitiated, isConsumable, price, currency, productName, originalTransactionId, callback));
			});
        }

        public void Payment(PaymentDTO paymentDto, Action<bool, Dictionary<String, System.Object>> callback = null)
        {
	        Payment(
		        paymentDto.isSubscription,
		        paymentDto.receipt,
		        paymentDto.userInitiated,
		        paymentDto.isConsumable,
		        paymentDto.price,
		        paymentDto.currency,
		        paymentDto.productName,
		        paymentDto.originalTransactionId,
		        callback
		        );
        }

        private IEnumerator PaymentRoutine(bool isSubscription, string receipt, bool userInitiated, bool isConsumable,
	        decimal? price = null, string currency = null, string productName = null, string originalTransactionId = null,
	        Action<bool, Dictionary<String, System.Object>> callback = null)
        {
            if (!RocGm.IsInitialized)
            {
                Log.Warning("Rocket GM is NOT initialized yet!", "PaymentRoutine");
                yield break;
            }

			Log.Debug(string.Format("Payment:" +
                                     "\nIsSubscription: {0}" +
                                     "\nPrice: {1}" +
                                     "\nCurrency: {2}" +
                                     "\nProductName: {3}" +
                                     "\nIsNewPayment: {4}" +
                                     "\nReceipt: {5}",
                isSubscription, price, currency, productName, userInitiated, receipt));

            WWWForm request = RocGm.CreateRequest();
            
			request.AddField("playerId", RocGm.PlayerId.ToString());
            request.AddField("isSubscription", isSubscription ? 1 : 0);
            request.AddField("receipt", receipt);
            request.AddField("isConsumable", isConsumable ? 1 : 0);

            if (price.HasValue)
            {
                request.AddField("price", price.Value.ToString(CultureInfo.InvariantCulture));
            }
            if (currency != null)
            {
                request.AddField("currency", currency);
            }
            if (productName != null)
            {
	            request.AddField("productName", productName);
            }
            request.AddField("userInitiated", userInitiated ? 1 : 0);

#if UNITY_IOS
            if (originalTransactionId != null)
            {
                request.AddField("originalTransactionId", originalTransactionId);
            }
#endif

            UnityWebRequest www = UnityWebRequest.Post(string.Format("{0}/payment", RocGm.RocGmTrackerUrl), request);

	        www.useHttpContinue = false;

			Log.Debug("Sending PAYMENT request");

            yield return www.SendWebRequest();

            if (www.responseCode != 200)
            {
                Log.Error(string.Format("Payment REJECTED! responseCode: {0} errorCode: {1}", www.responseCode, www.downloadHandler.text));
                if (callback != null) callback(false, null);
                yield break;
            }

            Dictionary<String, System.Object> result;
            if (!RocGmHelper.TryGetServerResult(www.downloadHandler.text, out result))
            {
                if (callback != null) callback(false, result);
                yield break;
            }

            if (callback != null) callback(true, result);

            Log.Info($"Payment APPROVED!");

            if (result != null)
            {
	            foreach (var (key, value) in result)
	            {
		            Log.Info($" Payment Result: {key}: {value}");
	            }

	            Log.Info(www.responseCode.ToString());
            }
        }

        public void SendRemotePlayerDataWhenPurchaseSuccessful()
        {
	        if(RocGm.GameSettings.RemotePlayerDataEnabled)
				_rocGmSessionReporting.ReportSessionWhenPurchaseSuccessful();
        }
#endregion

#region Attribution Tracking
        /// <summary>
        /// Update player's attribution tracking data
        /// </summary>
        public void UpdatePlayerAttribution()
        {
#if ROC_GM_APPSFLYER
            CoroutineController.StartCoroutine(UpdatePlayerAppsFlyerAttributionRoutine());
#elif ROC_GM_ADJUST
            CoroutineController.StartCoroutine(UpdatePlayerAdjustAttributionRoutine());
#endif
        }
#if ROC_GM_ADJUST
        private IEnumerator UpdatePlayerAdjustAttributionRoutine()
		{
			if (!RocGm.IsInitialized)
			{
				Log.Warning("Rocket GM is NOT initialized yet!", "UpdatePlayerAdjustAttributionRoutine");
				yield break;
			}

			if (RocGm.Config.IsAttributionUpdated)
            {
                Log.Error("Player Attribution data is ALREADY updated!");
                yield break;
            }

			WWWForm request = RocGm.CreateRequest();

			request.AddField("playerId", RocGm.PlayerId.ToString());

			if (!string.IsNullOrEmpty(RocGm.Config.AdjustId))
			{
				request.AddField("adjustId", RocGm.Config.AdjustId);
			}

			if (!string.IsNullOrEmpty(RocGm.Config.RefSource))
			{
				var refSource = RocGm.Config.GetNonEmptyRefSource();
				RefSource convertedRefSource = RocGm.RefSourceConverter.Convert(refSource);

				if (convertedRefSource == RefSource.None)
				{
					request.AddField("refSource", refSource);
				}
				else
				{
					request.AddField("refSource", (int) convertedRefSource);
				}
			}

			if (!string.IsNullOrEmpty(RocGm.Config.RefCampaign))
			{
				request.AddField("refCampaign", RocGm.Config.RefCampaign);
			}

			if (!string.IsNullOrEmpty(RocGm.Config.RefAdGroup))
			{
				request.AddField("refAdGroup", RocGm.Config.RefAdGroup);
			}

			if (!string.IsNullOrEmpty(RocGm.Config.RefCreative))
			{
				request.AddField("refCreative", RocGm.Config.RefCreative);
			}

			UnityWebRequest www = UnityWebRequest.Post(string.Format("{0}/updatePlayerAdjustAttributes", RocGm.RocGmTrackerUrl), request);

			www.chunkedTransfer = false;
			www.useHttpContinue = false;

			www.timeout = 10;

			Log.Debug("Sending UpdatePlayerAdjustAttributes request");

			yield return www.SendWebRequest();

			if (www.responseCode == 200)
			{
				Log.Debug("Player Attribution data is updated successfully");
                RocGm.Config.IsAttributionUpdated = true;
            }
			else
			{
				Log.Error("Player Attribution data could NOT be updated! Retrying...");
				CoroutineController.DoAfterGivenTime(_retryInterval, UpdatePlayerAttribution);
			}
		}        
#endif

#if ROC_GM_APPSFLYER
        private IEnumerator UpdatePlayerAppsFlyerAttributionRoutine()
        {
            if (!RocGm.IsInitialized)
            {
                Log.Warning("Rocket GM is NOT initialized yet!", "UpdatePlayerAdjustAttributionRoutine");
                yield break;
            }

            if (string.IsNullOrEmpty(RocGm.Config.ConversionData))
            {
                Log.Warning("Conversion Data is EMPTY!");
                yield break;
            }

            if (RocGm.Config.IsAttributionUpdated)
            {
                Log.Warning("Player Attribution data is ALREADY updated!");
                yield break;
            }

            WWWForm request = RocGm.CreateRequest();

            request.AddField("playerId", RocGm.PlayerId.ToString());
            request.AddField("conversionData", RocGm.Config.ConversionData);

            UnityWebRequest www = UnityWebRequest.Post(string.Format("{0}/updatePlayerAppsFlyerAttributes", RocGm.RocGmTrackerUrl), request);

            www.useHttpContinue = false;
            www.timeout = 10;

            Log.Debug("Sending UpdatePlayerAppsFlyerAttributes request");

            yield return www.SendWebRequest();

            if (www.responseCode != 200 || !RocGmHelper.TryGetServerResult(www.downloadHandler.text, out _appsFlyerAttributionParameters))
            {
                Log.Error(string.Format("UpdatePlayerAppsFlyerAttribution FAILED!\nresponseCode: {0}\nerror: {1}\ntext: {2}", www.responseCode, www.error, www.downloadHandler.text));
                CoroutineController.DoAfterGivenTime(_retryInterval, UpdatePlayerAttribution);
                yield break;
            }            

            RocGm.Config.IsAttributionUpdated = true;

            if(_appsFlyerAttributionParameters.ContainsKey("refSource"))
                RocGm.Config.RefSource = (string) _appsFlyerAttributionParameters["refSource"];

            if(_appsFlyerAttributionParameters.ContainsKey("refCampaign"))
                RocGm.Config.RefCampaign = (string) _appsFlyerAttributionParameters["refCampaign"];

            if(_appsFlyerAttributionParameters.ContainsKey("refAdGroup"))
                RocGm.Config.RefAdGroup = (string) _appsFlyerAttributionParameters["refAdGroup"];

            if(_appsFlyerAttributionParameters.ContainsKey("refCreative"))
                RocGm.Config.RefCreative = (string) _appsFlyerAttributionParameters["refCreative"];

            Log.Debug(string.Format("Player Attribution data is updated successfully:" +
                                    "\nRefSource:{0}" +
                                    "\nRefCampaign:{1}" +
                                    "\nRefAdGroup:{2}" +
                                    "\nRefCreative:{3}", 
                                    RocGm.Config.RefSource,
                                    RocGm.Config.RefCampaign,
                                    RocGm.Config.RefAdGroup,
                                    RocGm.Config.RefCreative));
        }
#endif
#endregion

#region Push Notification
        /// <summary>
        /// Update push token of user on server
        /// </summary>
        /// <param name="pushToken"></param>
        /// <param name="oneSignalId"></param>
        public void UpdatePushToken(string pushToken, [CanBeNull] string oneSignalId = null)
		{
			Log.Debug(string.Format("Token: {0}, OneSignalId: {1}", pushToken, oneSignalId), "UpdatePushToken");

			RocGm.DoAfterInitialized(() =>
			{
				CoroutineController.StartCoroutine(UpdatePushTokenRoutine(pushToken, oneSignalId));
			});
		}

		private IEnumerator UpdatePushTokenRoutine(string pushToken, [CanBeNull] string oneSignalId = null)
		{
		    if (string.IsNullOrEmpty(pushToken))
		    {
                Log.Error("Push token can NOT be empty");
                yield break;
		    }

		    if (RocGm.Config.IsPushTokenSent)
            {
				Log.Warning("PushToken is ALREADY updated!", "UpdatePushTokenRoutine");
                yield break;
            }

            if (!RocGm.IsInitialized)
            {
                Log.Warning("Rocket GM is NOT initialized yet!", "UpdatePushTokenRoutine");
                yield break;
            }

			WWWForm request = RocGm.CreateRequest();
			
			request.AddField("playerId", RocGm.PlayerId.ToString());
			request.AddField("pushToken", pushToken);

			UnityWebRequest www = UnityWebRequest.Post(string.Format("{0}/updatePlayerPushAttributes", RocGm.RocGmTrackerUrl), request);

			www.useHttpContinue = false;
			www.timeout = 10;

			Log.Debug(string.Format("Token: {0}", pushToken), "UpdatePushTokenRoutine");

			yield return www.SendWebRequest();

			Dictionary<string, object> parameters;

			if (www.responseCode != 200 || !RocGmHelper.TryGetServerResult(www.downloadHandler.text, out parameters))
			{
				Log.Error(string.Format("Push Attributes Update FAILED!\nresponseCode: {0}\nerror: {1}\ntext: {2}", www.responseCode, www.error, www.downloadHandler.text));
				CoroutineController.DoAfterGivenTime(_retryInterval, () => UpdatePushToken(pushToken, oneSignalId));
				yield break;
			}

			Log.Debug("Push Token is updated successfully");
			RocGm.Config.IsPushTokenSent = true;
		}

#endregion

#region AdvertisingId
        private void OnAdvertisingIdReceived()
        {
            RocGmControllerBase.DoAfterInitialized(() =>
            {
                if (RocGm.Config.IsFirstSession)
                    UpdatePlayerAdvertisingId(false);
                else
                    Log.Debug("Not First Session, Skip Advertiser Id Update");
            });
        }

        private void OnAuthorizationStatusChanged(IDFAAuthorizationStatus status)
        {
            RocGmIDFAController.AuthorizationStatusChanged -= OnAuthorizationStatusChanged;

            if (status == IDFAAuthorizationStatus.Authorized)
            {
                UpdatePlayerAdvertisingId(true);
            }

        }

        private void UpdatePlayerAdvertisingId(bool skipAdvertisingIdUpdated)
        {
            CoroutineController.StartCoroutine(UpdateAdvertisingIdCoroutine(skipAdvertisingIdUpdated));
        }

        private IEnumerator UpdateAdvertisingIdCoroutine(bool skipAdvertisingIdUpdated)
        {
            if (string.IsNullOrEmpty(RocGm.Config.AdvertisingId) || RocGm.Config.AdvertisingId == "00000000-0000-0000-0000-000000000000")
                yield break;

            if (!RocGm.IsInitialized)
            {
                Log.Warning("Rocket GM is NOT initialized yet!", "ReportAdvertisingIdCoroutine");
                RocGm.DoAfterInitialized(() => UpdatePlayerAdvertisingId(skipAdvertisingIdUpdated));
                yield break;
            }

            if (!skipAdvertisingIdUpdated && RocGm.Config.IsAdvertisingIdUpdated)
            {
                Log.Info("Advertising Id is ALREADY Sent!");
                yield break;
            }

            WWWForm request = RocGm.CreateRequest();

            request.AddField("playerId", RocGm.PlayerId.ToString());
            request.AddField("advertiserId", RocGm.Config.AdvertisingId);

            UnityWebRequest www = UnityWebRequest.Post(RocGm.RocGmTrackerUrl + "/updatePlayerAdvertiserId", request);

            www.useHttpContinue = false;

            www.timeout = 10;

            Log.Debug("Sending updatePlayerAdvertiserId request");

            yield return www.SendWebRequest();

            if (www.responseCode == 200)
            {
                Log.Debug("Player Advertising Id is updated successfully");
                RocGm.Config.IsAdvertisingIdUpdated = true;
            }
            else
            {
                Log.Error("Player Advertising Id could NOT be updated! Retrying...");
                CoroutineController.DoAfterGivenTime(_retryInterval, () => UpdatePlayerAdvertisingId(skipAdvertisingIdUpdated));
            }
        }

#endregion

        }
}
