using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using CodeStage.AntiCheat.ObscuredTypes;
using JetBrains.Annotations;
using JsonDiffPatchDotNet;
using Newtonsoft.Json.Linq;
using RocketCoroutine;
using RocketGM.Models;
using RocketGM.Models.UpdateSession;
using RocketGM.PanStats;
using RocketGM.Utils;
using RocketLog;
using RocketUtils;
using RocketUtils.Models;
using UnityEngine;
using UnityEngine.Networking;
using Debug = UnityEngine.Debug;
using Object = System.Object;

namespace RocketGM
{
    internal class RocGmSessionReporting
    {
        public static event Action PlayerDataIsDirty;
        public static event Action MaintenanceActivated;

        private bool _isSessionDataUpdateInProgress;

        private const int _retryInterval = 10;
        private const int _sendSessionDataInterval = 30;

        private int _rewardedCountInCurrentSession;
        private int _interstitialCountInCurrentSession;
        private int _bannerCountInCurrentSession;
        private double _estimatedBannerRevenue;
        private double _estimatedInterstitialRevenue;
        private double _estimatedRewardedRevenue;

        private const string _rawImpressionDataKey = "gmRawImpressionData";
        private const string _sentImpressionDataKey = "gmSentImpressionData";
        private const string _updatedSessionDataKeys = "gmUpdatedSessionDataKeys";
        private const string _durationKey = "gmSessionDuration";

        internal static ulong RemotePlayerDataDiffCount
        {
            get => ObscuredPrefs.GetULong("gmRemotePlayerDataDiffCount", 0);
            set => ObscuredPrefs.SetULong("gmRemotePlayerDataDiffCount", value);
        }

        private JsonDiffPatch _jsonDiffPatch = new JsonDiffPatch();

        private string RawImpressionData
        {
            get => PlayerPrefs.GetString(_rawImpressionDataKey, "{}");
            set => PlayerPrefs.SetString(_rawImpressionDataKey, value);
        }

        internal static string SentSessionData
        {
            get => PlayerPrefs.GetString(_sentImpressionDataKey, "{}");
            set => PlayerPrefs.SetString(_sentImpressionDataKey, value);
        }

        private string UpdatedSessionDataKeyString
        {
            get => PlayerPrefs.GetString(_updatedSessionDataKeys, "");
            set => PlayerPrefs.SetString(_updatedSessionDataKeys, value);
        }

        private int Duration
        {
            get => PlayerPrefs.GetInt(_durationKey, 0);
            set => PlayerPrefs.SetInt(_durationKey, value);
        }

        private IEnumerator _sendSessionDataInIntervalEnumerator;
        private const string _retrySessionDataKey = "RetrySessionData";
        private const string _retrySessionSuccessDataKey = "RetrySessionSuccess";

        private readonly Stopwatch _stopwatch = new Stopwatch();

        private readonly Dictionary<string, ImpressionData> _impressionDataCollection = new Dictionary<string, ImpressionData>();

        private readonly RocLog _log = new RocLog(nameof(RocGmSessionReporting), DebugLevels.Debug);

        internal RocGmSessionReporting()
        {
            RocGmIDFAController.IDFANativeUiAuthorizationInitiated += OnIDFANativeUiAuthorizationInitiated;
            RocGmIDFAController.AuthorizationStatusChanged += OnAuthorizationStatusChanged;

            RocGmControllerBase.DoAfterInitialized(() =>
            {
                RocGmControllerBase.PauseStateChanged += OnPauseStateChanged;
                RocGmControllerBase.FocusStateChanged += OnFocusStateChanged;

                _stopwatch.Start();

            });

            RocGm.DoAfterInitialized(() =>
            {
                if (!RocGm.GameSettings.LegacyMarketingEnabled)
                {
                    // Send unsent data on initialize.
                    if (!string.IsNullOrEmpty(SentSessionData))
                    {
                        ReportSession(0, false);
                    }
                    // Send unsent keys to server for approval on initialize.
                    else if (!string.IsNullOrEmpty(UpdatedSessionDataKeyString))
                    {
                        NotifyServerWithKeys();
                    }

                    //Start Session Data Cron
                    StartIntervalRoutine();
                }
            });
        }
        private void OnAuthorizationStatusChanged(IDFAAuthorizationStatus status)
        {
            HandleTimerAndReportSession(false);
        }
        private void OnIDFANativeUiAuthorizationInitiated(bool atStart, bool isInitial, int count)
        {
            if (atStart)
            {
                _stopwatch.Stop();
            }
        }

        ~RocGmSessionReporting()
        {
            RocGmIDFAController.IDFANativeUiAuthorizationInitiated -= OnIDFANativeUiAuthorizationInitiated;

            RocGmControllerBase.PauseStateChanged -= OnPauseStateChanged;
            RocGmControllerBase.FocusStateChanged -= OnFocusStateChanged;

            StopIntervalRoutine();
        }

        /// <summary>
        /// Increase AdWatch count depending on AdFormat.
        /// </summary>
        /// <param name="adFormat"></param>
        internal void SaveAdWatched(AdFormat adFormat)
        {
            ImpressionData impressionData = new ImpressionData
            {
                AdNetwork = "NaN",
                WatchCount = 1,
                AdRevenue = 0
            };

            switch (adFormat)
            {
                case AdFormat.Interstitial:
                    _interstitialCountInCurrentSession++;
#if ironsource_manager
                    impressionData.AdUnitFormat = "interstitial";
#endif
                    break;
                case AdFormat.RewardedVideo:
                    _rewardedCountInCurrentSession++;
#if ironsource_manager
                    impressionData.AdUnitFormat = "rewarded_video";
#endif
                    break;
                case AdFormat.Banner:
                    _bannerCountInCurrentSession++;
#if ironsource_manager
                    impressionData.AdUnitFormat = "banner";
#endif
                    break;
                default:
                    _log.Error("Enter a valid AdFormat! Current AdFormat: " + adFormat);
                    return;
            }

            SaveAdWatched(impressionData);

            _log.Debug("Ad Watch is count");
        }

        /// <summary>
        /// Save impression data got from Ad Network.
        /// </summary>
        /// <param name="impressionData"></param>
        internal void SaveAdWatched(ImpressionData impressionData)
        {
            string key = $"{impressionData.AdNetwork}_{impressionData.AdUnitFormat}";

            if (_impressionDataCollection.ContainsKey(key))
            {
                ImpressionData savedImpressionData = _impressionDataCollection[key];
                savedImpressionData.WatchCount += impressionData.WatchCount;
                savedImpressionData.AdRevenue += impressionData.AdRevenue;
                _impressionDataCollection[key] = savedImpressionData;
            }
            else
            {
                _impressionDataCollection.Add(key, impressionData);
            }

            RawImpressionData = JsonUtility.ToJson(new ImpressionDataContainer
            {
                impressionDataList = _impressionDataCollection.Values.ToList()
            });
        }

        /// <summary>
        /// Updates the estimated ad revenues for legacy marketing. This will only work if legacy marketing is enabled.
        /// </summary>
        /// <param name="revenue">The revenue value that will be added to related estimated revenue</param>
        /// <param name="adFormat">The format/type of the ad</param>
        internal void UpdateEstimatedAdRevenue(double revenue, AdFormat adFormat)
        {
            // Do not even try to update estimated revenues if legacy marketing is not enabled
            if (!RocGm.GameSettings.LegacyMarketingEnabled)
            {
                _log.Debug("Can not update estimated ad revenues because legacy marketing is not enabled!");
                return;
            }

            switch (adFormat)
            {
                case AdFormat.Banner:
                    _estimatedBannerRevenue += revenue;
                    break;
                case AdFormat.Interstitial:
                    _estimatedInterstitialRevenue += revenue;
                    break;
                case AdFormat.RewardedVideo:
                    _estimatedRewardedRevenue += revenue;
                    break;
            }

            _log.Debug($"Updated estimated ad revenue for {adFormat} by {revenue.ToString(CultureInfo.InvariantCulture)}");
        }

        /// <summary>
        /// Report session cron.
        /// </summary>
        /// <returns></returns>
        private IEnumerator SendSessionDataInInterval()
        {
            WaitForSeconds waitForSessionDataInterval = new WaitForSeconds(_sendSessionDataInterval);
            while (true)
            {
                yield return waitForSessionDataInterval;

                _log.Debug("Interval arrived. Send Session Data!");

                _stopwatch.Stop();
                var duration = Mathf.RoundToInt((float)_stopwatch.Elapsed.TotalSeconds);
                ReportSession(duration, true);
                RestartTimer();
            }
            // ReSharper disable once IteratorNeverReturns
        }

        /// <summary>
        /// Report Session on application pause
        /// </summary>
        /// <param name="paused"></param>
        private void OnPauseStateChanged(bool paused)
        {
#if ROC_AD
            if (_isAdOnProgress) return;
#endif
            HandleTimerAndReportSession(paused);
        }

        /// <summary>
        /// Report Session on application focus
        /// </summary>
        /// <param name="focused"></param>
        private void OnFocusStateChanged(bool focused)
        {
#if ROC_AD
            if (_isAdOnProgress) return;
#endif
            HandleTimerAndReportSession(!focused);
        }
        
#if ROC_AD
        private bool _isAdOnProgress = false;
        
        public void OnAdStarted()
        {
            _isAdOnProgress = true;
            // Lost focus
            HandleTimerAndReportSession(true);
        }
        public void OnAdFinished()
        {
            _isAdOnProgress = false;
            // Gained focus
            HandleTimerAndReportSession(false);
        }
#endif

        private void HandleTimerAndReportSession(bool shouldStop)
        {
            if (shouldStop && _stopwatch.IsRunning)
            {
                _stopwatch.Stop();
                var duration = Mathf.RoundToInt((float)_stopwatch.Elapsed.TotalSeconds);

                ReportSession(duration, true);
            }
            else if (!shouldStop && !_stopwatch.IsRunning)
            {
                RestartTimer();
            }
        }

        internal void ReportSessionWhenPurchaseSuccessful()
        {
            ReportSession(1, true);
        }
        
        /// <summary>
        /// Wrapper report session method which supports both legacy marketing and impression based marketing
        /// </summary>
        /// <param name="duration"></param>
        /// <param name="updateDuration"></param>
        private void ReportSession(int duration, bool updateDuration)
        {
            if (updateDuration)
            {
                Duration += duration;
            }

            if (RocGm.IsInitialized)
            {
                if (RocGm.GameSettings.LegacyMarketingEnabled)
                {
                    ReportSessionForLegacyMarketing();
                }
                else
                {
                    if (!_isSessionDataUpdateInProgress)
                    {
                        _isSessionDataUpdateInProgress = true;
                        ReportSessionForImpressionBasedMarketing(updateDuration);
                    }
                    else
                    {
                        _log.Warning("Task Is Already Running!");
                    }
                }
            }
        }

        /// <summary>
        /// Report session which contains Duration and Ad Watch count
        /// Important! Supports older clients which uses legacy marketing!
        /// </summary>
        private void ReportSessionForLegacyMarketing()
        {
            if (!RocGm.IsInitialized)
            {
                _log.Warning("Rocket GM is NOT initialized yet!");
                return;
            }

            var duration = Duration;

            if (duration < 1)
            {
                _log.Error("Duration can NOT be lower than 1!");
                return;
            }

            WWWForm request = RocGm.CreateRequest();

            request.AddField("playerId", RocGm.PlayerId.ToString());
            request.AddField("duration", duration);
            request.AddField("rewardedWatched", _rewardedCountInCurrentSession);
            request.AddField("interstitialWatched", _interstitialCountInCurrentSession);
            request.AddField("bannerWatched", _bannerCountInCurrentSession);
            request.AddField("estBannerRevenue", (int)_estimatedBannerRevenue);
            request.AddField("estInterstitialRevenue", (int)_estimatedInterstitialRevenue);
            request.AddField("estRewardedRevenue", (int)_estimatedRewardedRevenue);

            UnityWebRequest www = UnityWebRequest.Post(string.Format("{0}/updateSession", RocGm.RocGmTrackerUrl), request);

            www.useHttpContinue = false;

            www.SendWebRequest();

            _rewardedCountInCurrentSession = 0;
            _interstitialCountInCurrentSession = 0;
            _bannerCountInCurrentSession = 0;

            _estimatedBannerRevenue = 0.0;
            _estimatedInterstitialRevenue = 0.0;
            _estimatedRewardedRevenue = 0.0;

            ClearImpressionData();
        }

        /// <summary>
        /// Report session which contains Duration and Impression Data.
        /// </summary>
        /// <param name="updateDuration"></param>
        /// <returns></returns>
        private async Task ReportSessionForImpressionBasedMarketing(bool updateDuration)
        {
            try
            {
                if (!RocGm.IsInitialized)
                {
                    _log.Warning("Rocket GM is NOT initialized yet!");
                    _isSessionDataUpdateInProgress = false;
                    return;
                }

                var duration = Duration;

                if (updateDuration && duration < 1)
                {
                    _log.Warning("Duration can NOT be lower than 1!");
                    _isSessionDataUpdateInProgress = false;
                    return;
                }

                StopIntervalRoutine();

                string sentSessionData = SentSessionData;
                var sentSessionDataContainer = JsonUtility.FromJson<SentSessionDataContainer>(sentSessionData);

                //Add new session data, if not empty
                string rawImpressionData = RawImpressionData;
                var now = DateTime.UtcNow;
                if (RocGm.FetchServerTimeStatus == RocGm.ServerTimeStatus.Initialized)
                {
                    now = RocGm.ServerTimeNow;
                }

                string playerData = "";
                string playerDiffData = "";

                if (updateDuration && RocGm.GameSettings.RemotePlayerDataEnabled)
                {
                    if (PlayerStats.TryGetPlayerDataAsJsonString(true, out playerData))
                    {
                        playerDiffData = _jsonDiffPatch.Diff(RocGm.RemotePlayerData, playerData);
                        if (playerDiffData != null)
                        {
                            RemotePlayerDataDiffCount++;
                        }
                        else
                        {
                            playerDiffData = "";
                        }
                    }
                }

                if (duration > 0 || rawImpressionData != "{}" || playerDiffData != "")
                {
                    SentSessionData toBeSentSessionData = new SentSessionData
                    {
                        Key = RocGm.PlayerId + "_" + ((DateTimeOffset)now).ToUnixTimeSeconds(),
                        Duration = duration,
                        ImpressionData = rawImpressionData,
                    };

                    if (playerDiffData != "")
                    {
                        //TODO - Ahmet: test for no data
                        toBeSentSessionData.RemotePlayerDataDiff = playerDiffData;
                        toBeSentSessionData.RemotePlayerDataCount = RemotePlayerDataDiffCount;
                        RocGm.RemotePlayerData = playerData;
                    }

                    _log.Debug($"NewSessionData: Duration: {duration}, ImpressionData: {rawImpressionData}, RemotePlayerDataDiff: {playerDiffData}, RemotePlayerDataCount: {RemotePlayerDataDiffCount}");

                    sentSessionDataContainer.SentImpressionDataList.Add(toBeSentSessionData);
                    sentSessionData = JsonUtility.ToJson(sentSessionDataContainer);
                    SentSessionData = sentSessionData;

                    ClearImpressionData();
                }

                List<string> sentKeys = new List<string>();

                foreach (SentSessionData sessionData in sentSessionDataContainer.SentImpressionDataList)
                {
                    sentKeys.Add(sessionData.Key);
                }

                if (sentKeys.Count == 0)
                {
                    _isSessionDataUpdateInProgress = false;
                    return;
                }

                _log.Debug(string.Format("Sending Session Data:\n{0}", sentSessionData));
                WWWForm request = RocGm.CreateRequest();
                request.AddField("playerId", RocGm.PlayerId.ToString());

                if (RocGm.Config.IsEncryptedMessagingActive)
                {
                    request.AddField("sessionData",
                        RocGmSecurity.Encrypt(sentSessionData, RocGm.Config.SecurityKey));
                }
                else
                {
                    request.AddField("sessionData", sentSessionData);
                }

                #if ironsource_manager
                request.AddField("mediation", 2);
                #endif

                using var www = UnityWebRequest.Post(string.Format("{0}/updateSessionData", RocGm.RocGmTrackerUrl), request);
                www.useHttpContinue = false;
                www.timeout = 10;

                await www.SendWebRequest();

                if (www.responseCode != 200)
                {
                    _log.Warning(string.Format("ReportSessionForNewMarketing FAILED!\nresponseCode: {0}\nerror: {1}\ntext: {2}", www.responseCode, www.error, www.downloadHandler.text));
                    CoroutineController.DoAfterGivenTime(_retryInterval, () => ReportSession(duration, false), _retrySessionDataKey);
                    _isSessionDataUpdateInProgress = false;
                    www.Dispose();
                    return;
                }

                //Notify server with updated keys.
                foreach (var updatedKey in sentKeys)
                {
                    var dataList = sentSessionDataContainer.SentImpressionDataList;
                    for (var index = dataList.Count - 1; index >= 0; index--)
                    {
                        SentSessionData sessionData = sentSessionDataContainer.SentImpressionDataList[index];
                        if (sessionData.Key == updatedKey)
                        {
                            dataList.Remove(sessionData);
                            break;
                        }
                    }
                }

                SentSessionData = sentSessionDataContainer.SentImpressionDataList.Count == 0 ? "{}" : JsonUtility.ToJson(sentSessionDataContainer);

                if (string.IsNullOrEmpty(UpdatedSessionDataKeyString))
                {
                    UpdatedSessionDataKeyString = string.Join(",", sentKeys);
                }
                else
                {
                    UpdatedSessionDataKeyString += "," + string.Join(",", sentKeys);
                }


                NotifyServerWithKeys();

                RocGmHelper.TryGetServerResult(www.downloadHandler.text, out var parameters);

                if (parameters.ContainsKey(RocGm.MaintenanceKey) && parameters[RocGm.MaintenanceKey] is bool isMaintenance && isMaintenance)
                {
                    MaintenanceActivated?.Invoke();
                }

                if (parameters.ContainsKey(RocGm.PlayerDataIsDirtyKey))
                {
                    PlayerDataIsDirty?.Invoke();
                }

                StartIntervalRoutine();

                _isSessionDataUpdateInProgress = false;
            }
            catch (Exception e)
            {
                _log.Exception(e.Message);
                _isSessionDataUpdateInProgress = false;
            }

        }


        private void NotifyServerWithKeys()
        {
            if (string.IsNullOrEmpty(UpdatedSessionDataKeyString))
            {
                _log.Warning("No keys to notify server!", "NotifyServer");
                return;
            }

            CoroutineController.StartCoroutine(NotifyServerWithKeysRoutine());
        }

        private IEnumerator NotifyServerWithKeysRoutine()
        {
            if (!RocGm.IsInitialized)
            {
                _log.Warning("Rocket GM is NOT initialized yet!");
                yield break;
            }

            var updatedSessionDataKeyString = UpdatedSessionDataKeyString;

            WWWForm request = RocGm.CreateRequest();
            request.AddField("keys", updatedSessionDataKeyString);

            _log.Debug("Sending UpdateSessionDataSuccess with: " + updatedSessionDataKeyString);

            using var www = UnityWebRequest.Post($"{RocGm.RocGmTrackerUrl}/updateSessionDataSuccess", request);
            www.useHttpContinue = false;
            www.timeout = 10;

            yield return www.SendWebRequest();

            if (www.responseCode != 200)
            {
                _log.Warning(string.Format("NotifyServer FAILED!\nresponseCode: {0}\nerror: {1}\ntext: {2}", www.responseCode, www.error, www.downloadHandler.text));
                CoroutineController.DoAfterGivenTime(_retryInterval, NotifyServerWithKeys, _retrySessionSuccessDataKey);
                www.Dispose();
                yield break;
            }

            _log.Debug("UpdateSessionDataSuccess successful");
            www.Dispose();
            UpdatedSessionDataKeyString = "";
        }

        private void RestartTimer()
        {
            _stopwatch.Reset();
            _stopwatch.Start();
        }

        private void StartIntervalRoutine()
        {
            StopIntervalRoutine();

            CoroutineController.StartCoroutine(_sendSessionDataInIntervalEnumerator = SendSessionDataInInterval());
        }

        private void StopIntervalRoutine()
        {
            if (!CoroutineController.IsCoroutineRunning(_sendSessionDataInIntervalEnumerator))
                return;

            CoroutineController.StopCoroutine(_sendSessionDataInIntervalEnumerator);
            _sendSessionDataInIntervalEnumerator = null;
        }

        private void ClearImpressionData()
        {
            _impressionDataCollection.Clear();
            RawImpressionData = "{}";
            Duration = 0;
        }

        public static SentSessionDataContainer GetSentSessionDataContainer()
        {
            return JsonUtility.FromJson<SentSessionDataContainer>(SentSessionData);
        }
    }
}
